"""
XLSX Data Plugin
****************

This plugin loads data from xlsx tables and transforms it in a dictionary
supported by N2G diagram plugins. Using ``from_dict`` method, this plugin
loads data into diagram plugin.

Guidelines and Limitations
--------------------------

* `openpyxl <https://pypi.org/project/openpyxl/>`_ >= 3.0.0 library need to be installed: ``pip install openpyxl``
* Nodes and links tabs' first row must contain headers
* nodes tab should have at least ``id`` header, other headers should comply with ``from_dict`` method attributes
  or simply ignored
* links tab should have at least ``source`` and ``target`` headers, other headers should comply with ``from_dict``
  method attributes or simply ignored

Sample Usage
------------

Code to invoke ``xlsx_data``::

    from N2G import drawio_diagram
    from N2G import xlsx_data

    drawio_drawing = drawio_diagram()

    xlsx_data(
        drawio_drawing,
        "./Data/nodes_and_links_data.xlsx",
        node_tabs="nodes",
        link_tabs="links"
    )

    drawio_drawing.layout(algo="kk")
    drawio_drawing.dump_file(filename="diagram.drawio", folder="./Output/")

Where ``nodes_and_links_data.xlsx`` content for nodes tab::

    id   pic    label    bottom_label    top_label    description
    r1          r1       core            1.1.1.1      Core Router
    r2          r2       core            2.2.2.2      Core Router
    r3          r3       edge            3.3.3.3      Edge Router

for links tab::

    source    src_label    label    target    trgt_label    description
    r1        Gi1/1        DF-10Km  r2        Gi3/4         DF link between R1 and R2
    r3        10GE2/1/1    DF-32Km  r2        Ten1/1        DF link between R3 and R2

Support available to translate headers to comply with N2G diagram modules
``from_dict`` or ``from_list`` methods through the use of ``node_headers_map`` and
``link_headers_map``.  For instance consider this table::

    # nodes tab:
    hostname   lo0_ip    bgp_asn
    r1         1.1.1.1   65123
    r2         1.1.1.2   65123
    r3         1.1.1.3   65123

    # links tab:
    device:a    interface:a    label    device:b  interface:b
    r1          Gi1/1          DF-10Km  r2        Gi3/4
    r3          10GE2/1/1      DF-32Km  r2        Ten1/1

If ``node_headers_map`` is::

    node_headers_map = {
        "id": ["device", "hostname"],
        "tob_label": ["lo0_ip"],
        "bottom_label": ["bgp_asn"]
    }

And ``link_headers_map`` is::

    link_headers_map = {
        "source": ["device:a", "hostname:a"],
        "target": ["device:b", "hostname:b"],
        "src_label": ["interface:a", "ip:a"],
        "trgt_label": ["interface:b", "ip:b"]
    }

Above table will be transformed to::

    # nodes tab:
    id         tob_label bottom_label
    r1         1.1.1.1   65123
    r2         1.1.1.2   65123
    r3         1.1.1.3   65123

    # links tab:
    source      src_label      label    target    trgt_label
    r1          Gi1/1          DF-10Km  r2        Gi3/4
    r3          10GE2/1/1      DF-32Km  r2        Ten1/1

API Reference
-------------

.. autofunction:: N2G.plugins.data.xlsx_data.xlsx_data
"""

import logging
import traceback

# initiate logging
log = logging.getLogger(__name__)

try:
    from openpyxl import load_workbook
except ImportError:
    log.debug("N2G xlsx data plugin failed to import openpyxl module")


def translate_headers(headers, translate_dict):
    """
    Helper function to perform headers translation.

    :param headers: (list) list of headers to translate
    :param translate_dict: (dict) map of headers to translate to

    For example if ``headers`` are::

        headers = ["device:a", "interface:a", "label", "device:b", "interface:b"]

    and ``translate_dict`` is::

        translate_dict = {
            "source": ["device:a", "hostname:a"],
            "target": ["device:b", "hostname:b"],
            "src_label": ["src_label", "interface:a", "ip:a"],
            "trgt_label": ["trgt_label", "interface:b", "ip:b"]
        }

    this function will transform ``headers`` list in::

        headers = ["source", "src_label", "label", "target", "trgt_label"]
    """
    for index, header in enumerate(headers):
        if header in translate_dict:
            continue
        for translate_header, headers_list in translate_dict.items():
            if header in headers_list:
                headers[index] = translate_header
                break


def xlsx_data(
    drawing,
    data,
    node_tabs=None,
    link_tabs=None,
    node_headers_map=None,
    link_headers_map=None,
):
    """
    Function to load data from XLSX file and add it to diagram using
    ``from_dict`` method.

    :param drawing: N2G drawing module object
    :param data: (str) OS path to xlsx file to load
    :param node_tabs: (list) list of tabs with nodes data, default ``["nodes"]``
    :param link_tabs: (list) list of tabs with links data, default ``["links"]``
    :param node_headers_map: (dict) dictionary to use to translate node tabs headers
    :param link_headers_map: (dict) dictionary to use to translate link tabs headers
    :return: ``True`` on success and ``False`` on failure to load data
    """
    node_tabs = node_tabs or ["nodes"]
    link_tabs = link_tabs or ["links"]
    node_headers_map = node_headers_map or {"id": ["device", "hostname"]}
    link_headers_map = link_headers_map or {
        "source": ["device:a", "hostname:a"],
        "target": ["device:b", "hostname:b"],
        "src_label": ["interface:a", "ip:a"],
        "trgt_label": ["interface:b", "ip:b"],
    }
    wb = load_workbook(data, data_only=True, read_only=True)

    graph_dict = {"nodes": [], "links": []}

    try:
        # form graph dictionary
        for node_tab in node_tabs:
            node_headers = []
            for row in wb[node_tab].iter_rows(values_only=True):
                if not node_headers:
                    node_headers = [str(i).strip() for i in row]
                    translate_headers(node_headers, node_headers_map)
                else:
                    graph_dict["nodes"].append(
                        dict(zip(node_headers, [c if c else "" for c in row]))
                    )

        for link_tab in link_tabs:
            link_headers = []
            for row in wb[link_tab].iter_rows(values_only=True):
                if not link_headers:
                    link_headers = [str(i).strip() for i in row]
                    translate_headers(link_headers, link_headers_map)
                else:
                    graph_dict["links"].append(
                        dict(zip(link_headers, [c if c else "" for c in row]))
                    )

        # add data to graph
        drawing.from_dict(graph_dict, diagram_name="Page-1")
    except:
        log.error("N2G:xlsx_data, failed: {}".format(traceback.format_exc()))
        return False

    # clean up
    wb.close()
    del graph_dict, wb

    return True



"""
CLI ISIS LSDB Data Plugin
*************************

This module designed to process ISIS Link State Database (LSDB) of network devices
CLI output and make diagram out of it.

Show commands output from devices parsed using TTP Templates into a dictionary structure.

After parsing, results processed further to form a dictionary of nodes and links keyed
by unique nodes and links identifiers, dictionary values are nodes dictionaries and for links
it is a list of dictionaries of links between pair of nodes. For nodes ISIS RID
used as a unique ID, for links it is sorted tuple of ``source``, ``target`` and ``label``
keys' values. This structure helps to eliminate duplicates.

Next step is post processing, such as packing links between nodes or IP lookups.

Last step is to populate N2G drawing with new nodes and links using ``from_dict`` method.

Features Supported
------------------

**Support matrix**

+---------------+------------+
| Platform      | ISIS       |
| Name          | LSDB       |
+===============+============+
| cisco_ios     |     ---    |
+---------------+------------+
| cisco_xr      |     YES    |
+---------------+------------+
| cisco_nxos    |     ---    |
+---------------+------------+
| huawei        |     ---    |
+---------------+------------+

Required Commands output
------------------------

cisco_xr:

* ``show isis database verbose`` - mandatory output, used to parse ISIS LSDB content

Sample usage
------------

Code to populate yEd diagram object with ISIS LSDB sourced nodes and links::

    from N2G import yed_diagram as create_yed_diagram
    from N2G import cli_isis_data

    isis_lsdb_data = {"cisco_xr": ['''
    RP/0/RP0/CPU0:ROUTER-X1#show isis database verbose

    IS-IS 1234 (Level-2) Link State Database
    LSPID                 LSP Seq Num  LSP Checksum  LSP Holdtime/Rcvd  ATT/P/OL
    ROUTER-X1.00-00      * 0x00000832   0x74bc        64943/*            0/0/0
    Auth:           Algorithm HMAC-MD5, Length: 17
    Area Address:   49.1234
    NLPID:          0xcc
    Router ID:      10.211.1.1
    Hostname:       ROUTER-X1
    Metric: 0          IP-Extended 10.211.1.1/32
        Prefix Attribute Flags: X:1 R:0 N:0 E:0 A:0
    Metric: 16777214   IS-Extended ROUTER-X2.00
        Local Interface ID: 9, Remote Interface ID: 50
        Interface IP Address: 10.123.0.17
        Neighbor IP Address: 10.123.0.18
        Affinity: 0x00000000
        Physical BW: 10000000 kbits/sec
        Reservable Global pool BW: 0 kbits/sec
        Global Pool BW Unreserved:
        [0]: 0        kbits/sec          [1]: 0        kbits/sec
        [2]: 0        kbits/sec          [3]: 0        kbits/sec
        [4]: 0        kbits/sec          [5]: 0        kbits/sec
        [6]: 0        kbits/sec          [7]: 0        kbits/sec
        Admin. Weight: 1000
        Ext Admin Group: Length: 32
        0x00000000   0x00000000
        0x00000000   0x00000000
        0x00000000   0x00000000
        0x00000000   0x00000000
        Physical BW: 10000000 kbits/sec
    Metric: 802        IS-Extended ROUTER-X5.00
        Local Interface ID: 7, Remote Interface ID: 53
        Interface IP Address: 10.123.0.25
        Neighbor IP Address: 10.123.0.26
        Affinity: 0x00000000
        Physical BW: 10000000 kbits/sec
        Reservable Global pool BW: 0 kbits/sec
        Global Pool BW Unreserved:
        [0]: 0        kbits/sec          [1]: 0        kbits/sec
        [2]: 0        kbits/sec          [3]: 0        kbits/sec
        [4]: 0        kbits/sec          [5]: 0        kbits/sec
        [6]: 0        kbits/sec          [7]: 0        kbits/sec
        Admin. Weight: 802
        Ext Admin Group: Length: 32
        0x00000000   0x00000000
        0x00000000   0x00000000
        0x00000000   0x00000000
        0x00000000   0x00000000
        Physical BW: 10000000 kbits/sec
    ROUTER-X2.00-00        0x00000826   0x4390        65258/65535        0/0/0
    Auth:           Algorithm HMAC-MD5, Length: 17
    Area Address:   49.1234
    NLPID:          0xcc
    Router ID:      10.211.1.2
    Hostname:       ROUTER-X2
    Metric: 0          IP-Extended 10.211.1.2/32
        Prefix Attribute Flags: X:1 R:0 N:0 E:0 A:0
    Metric: 301        IS-Extended ROUTER-X6.00
        Local Interface ID: 48, Remote Interface ID: 53
        Interface IP Address: 10.123.0.33
        Neighbor IP Address: 10.123.0.34
        Affinity: 0x00000000
        Physical BW: 10000000 kbits/sec
        Reservable Global pool BW: 0 kbits/sec
        Global Pool BW Unreserved:
        [0]: 0        kbits/sec          [1]: 0        kbits/sec
        [2]: 0        kbits/sec          [3]: 0        kbits/sec
        [4]: 0        kbits/sec          [5]: 0        kbits/sec
        [6]: 0        kbits/sec          [7]: 0        kbits/sec
        Admin. Weight: 301
        Ext Admin Group: Length: 32
        0x00000000   0x00000000
        0x00000000   0x00000000
        0x00000000   0x00000000
        0x00000000   0x00000000
        Physical BW: 10000000 kbits/sec
    Metric: 16777214   IS-Extended ROUTER-X1.00
        Local Interface ID: 50, Remote Interface ID: 9
        Interface IP Address: 10.123.0.18
        Neighbor IP Address: 10.123.0.17
        Affinity: 0x00000000
        Physical BW: 10000000 kbits/sec
        Reservable Global pool BW: 0 kbits/sec
        Global Pool BW Unreserved:
        [0]: 0        kbits/sec          [1]: 0        kbits/sec
        [2]: 0        kbits/sec          [3]: 0        kbits/sec
        [4]: 0        kbits/sec          [5]: 0        kbits/sec
        [6]: 0        kbits/sec          [7]: 0        kbits/sec
        Admin. Weight: 1000
        Ext Admin Group: Length: 32
        0x00000000   0x00000000
        0x00000000   0x00000000
        0x00000000   0x00000000
        0x00000000   0x00000000
        Physical BW: 10000000 kbits/sec

    Total Level-2 LSP count: 2     Local Level-2 LSP count: 1
    RP/0/RP0/CPU0:ROUTER-X1#
        ''']
    }

    drawing = create_yed_diagram()
    drawer = cli_isis_data(drawing)
    drawer.work(isis_lsdb_data)
    drawing.dump_file()

API Reference
-------------

.. autoclass:: N2G.plugins.data.cli_isis_data.cli_isis_data
   :members:
"""
import logging
import json
import os
import csv
import ipaddress
from fnmatch import fnmatchcase
from N2G.utils import merge_dict

try:
    from ttp import ttp
    from ttp_templates import get_template

    HAS_TTP = True
except ImportError:
    HAS_TTP = False

# initiate logging
log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# Main class:
# -----------------------------------------------------------------------------


class cli_isis_data:
    """
    Main class to instantiate ISIS LSDB Data Plugin object.

    :param drawing: (obj) N2G Diagram object
    :param ttp_vars: (dict) Dictionary to use as vars attribute while instantiating TTP parser object
    :param platforms: (list) - list of platform names to process e.g. ``cisco_ios``, ``cisco_xr`` etc, default is ``_all_``
    :param ip_lookup_data: (dict or str) IP Lookup dictionary or OS path to CSV file
    :param add_connected: (bool) if True, will add connected subnets as nodes, default is False
    :param ptp_filter: (list) list of glob patterns to filter point-to-point links based on link IP
    :param add_data: (bool) if True (default) adds data information to nodes and links

    ``ip_lookup_data`` dictionary must be keyed by ISSI RID IP address, with values
    being dictionary which must contain ``hostname`` key with optional additional keys
    to use for N2G diagram module node, e.g. ``label``, ``top_label``, ``bottom_label``,
    ``interface``etc. If ``ip_lookup_data`` is an OS path to CSV file, that file's first
    column header must be ``ip`` , file must contain ``hostname`` column, other columns
    values set to N2G diagram module node attributes, e.g. ``label``, ``top_label``,
    ``bottom_label``, ``interface`` etc.

    If lookup data contains ``interface`` key, it will be added to link label.

    Sample ip_lookup_data dictionary::

        {
            "1.1.1.1": {
                "hostname": "router-1",
                "bottom_label": "1 St address, City X",
                "interface": "Gi1"
            }
        }

    Sample ip_lookup_data CSV file::

        ip,hostname,bottom_label,interface
        1.1.1.1,router-1,"1 St address, City X",Gi1
    """

    def __init__(
        self,
        drawing,
        ttp_vars: dict = None,
        ip_lookup_data: dict = None,
        add_connected: bool = False,
        ptp_filter: list = None,
        add_data: bool = True,
        platforms: list = None,
    ):
        self.ttp_vars = ttp_vars or {}
        self.drawing = drawing
        self.drawing.node_duplicates = "update"
        self.add_connected = add_connected
        self.ptp_filter = ptp_filter or []
        self.add_data = add_data
        self.platforms = platforms or ["_all_"]
        self.parsed_data = {}
        self.nodes_dict = {}
        self.links_dict = {}
        self.graph_dict = {"nodes": [], "links": []}
        self.ip_lookup_data = ip_lookup_data or {}
        self._load_ip_lookup_data()

    def _load_ip_lookup_data(self) -> None:
        """
        Helper function to load CSV table in a dictionary keyed
        by values in first column. This dictionary further used
        to perform IP to node details lookup for ISIS router ID.
        """
        # load lookup data
        if self.ip_lookup_data and isinstance(self.ip_lookup_data, str):
            with open(self.ip_lookup_data) as f:
                reader = csv.DictReader(f)
                self.ip_lookup_data = {r["ip"]: r for r in reader if "ip" in r}

    def work(self, data):
        """
        Method to parse text data and add nodes and links to N2G drawing.

        :param data: (dict or str) dictionary keyed by platform name or OS path
          string to directories with text files

        If data is dictionary, keys must correspond to "Platform" column in
        `Features Supported`_ section table, values are lists of text items to
        process.

        Data dictionary sample::

            data = {
                "cisco_ios" : ["h1", "h2"],
                "cisco_xr": ["h3", "h4"],
                "cisco_nxos": ["h5", "h6"],
                ...etc...
            }

        Where ``hX`` device's show commands output.

        If data is an OS path directory string, child directories' names must correspond
        to **Platform** column in `Features Supported`_ section table. Each child directory
        should contain text files with show commands output for each device, names of files
        are arbitrary, but output should contain device prompt to extract device hostname.

        Directories structure sample::

            ├───folder_with_data
                ├───cisco_ios
                │       switch1.txt
                │       switch2.txt
                └───cisco_nxos
                        nxos_switch_1.txt
                        nxos_switch_2.txt

        To point N2G to above location ``data`` attribute string can be ``/var/data/n2g/folder_with_data/``
        """
        self._parse(data)
        self._form_base_graph_dict()
        self._pack_links()
        if self.ip_lookup_data:
            self._lookup_rid()
            self._lookup_ip_interfaces()
        self._update_drawing()

    def _make_hash_tuple(self, data: dict) -> tuple:
        """
        Helper function to form Edge tuple to use as a hash to
        identify all the links between same pair of nodes using
        ``source``, ``target`` and ``label`` keys' values.

        :param data: (dict) link dictionary with source, target and label keys
        """
        return tuple(sorted([data["source"], data["target"], data.get("label", "")]))

    def _parse(self, data: [dict, str]) -> None:
        """
        Method to parse data using TTP Templates

        :param data: (dict or str) dictionary of data items or OS path to folders with data to parse
        :return: None
        """
        if not HAS_TTP:
            raise ModuleNotFoundError(
                "N2G:cli_isis_data failed importing TTP, is it installed?"
            )
        parser = ttp(vars=self.ttp_vars, log_level="ERROR")
        # process data dictionary
        if isinstance(data, dict):
            for platform_name, text_list in data.items():
                if (
                    "_all_" not in self.platforms
                    and not platform_name in self.platforms
                ):
                    continue
                ttp_template = get_template(
                    misc="N2G/cli_isis_data/{}.txt".format(platform_name)
                )
                parser.add_template(template=ttp_template, template_name=platform_name)
                for item in text_list:
                    parser.add_input(item, template_name=platform_name)
        # process directories at OS path
        elif isinstance(data, str):
            # get all sub-folders and load respective templates
            with os.scandir(data) as dirs:
                for entry in dirs:
                    if entry.is_dir():
                        platform_name = entry.name
                        if (
                            "_all_" not in self.platforms
                            and not platform_name in self.platforms
                        ):
                            continue
                        ttp_template = get_template(
                            misc="N2G/cli_isis_data/{}.txt".format(platform_name)
                        )
                        parser.add_template(
                            template=ttp_template, template_name=platform_name
                        )
                        parser.add_input(
                            data=os.path.abspath(entry), template_name=platform_name
                        )
        else:
            raise TypeError(
                "Expecting dictionary or string, but '{}' given".format(type(data))
            )
        parser.parse(one=True)
        self.parsed_data = parser.result(structure="flat_list")
        # import pprint; pprint.pprint(self.parsed_data, width = 100)

    def _process_lsp(self, lsp: dict, isis_pid: str, hostname: str) -> None:
        """
        Method to process ISIS LSP dictionary extracting node, links and subnets.

        :param lsp: (dict) LSP dictionary
        :param isis_pid: (str) ISIS Process ID string
        :param hostname: device hostname LSP belongs to
        """
        # make node out of router LSP
        node = {"id": hostname, "label": hostname, "bottom_label": "Node"}
        if lsp.get("rid") or lsp.get("rid_v6"):
            node["top_label"] = lsp.get("rid") or lsp.get("rid_v6")
        self._add_node(node, node_data=lsp)
        # go over links
        for link in lsp.get("links", []):
            # ignore ISIS links based on IP addresses
            if any([fnmatchcase(link.get("local_ip", ""), p) for p in self.ptp_filter]):
                continue
            self._add_link(
                link={
                    "source": hostname,
                    "src_label": "{}:{}".format(
                        link.get("local_ip", link.get("local_intf_id")), link["metric"]
                    ),
                    "label": "{}:{}".format(
                        isis_pid, lsp["level"].replace("Level-", "L")
                    ),
                    "target": link["peer_name"],
                    "local_intf_id": link.get("local_intf_id"),
                    "peer_intf_id": link.get("peer_intf_id"),
                    "peer_ip": link.get("peer_ip"),
                    "local_ip": link.get("local_ip"),
                },
                link_data={hostname: {"isis_pid": isis_pid, **link}},
            )
        # go over connected subnets
        if self.add_connected:
            for network in lsp.get("networks", []):
                self._add_node(
                    node={
                        "id": network["network"],
                        "label": network["network"],
                        "bottom_label": "Subnet",
                    }
                )
                self._add_link(
                    link={
                        "source": hostname,
                        "src_label": "M:{}".format(network["metric"]),
                        "label": lsp["level"],
                        "target": network["network"],
                    }
                )

    def _form_base_graph_dict(self) -> None:
        for device in self.parsed_data:
            # go over all ISIS processes on the box
            for isis_pid, isis_data in device.get("isis_processes", {}).items():
                # process LSP
                for hostname, lsps in isis_data.items():
                    for lsp in lsps:
                        self._process_lsp(lsp, isis_pid, hostname)

    def _add_node(self, node: dict, node_data: dict = {}) -> None:
        # add new node
        if not node["id"] in self.nodes_dict:
            if node_data and self.add_data:
                node["description"] = json.dumps(
                    node_data, sort_keys=True, indent=4, separators=(",", ": ")
                )
            self.nodes_dict[node["id"]] = node
        # merge existing and new nodes
        else:
            node_ref = self.nodes_dict[node["id"]]
            merge_dict(node_ref, node)
            if node_data and self.add_data:
                data_ref = json.loads(node_ref.get("description", "{}"))
                merge_dict(data_ref, node_data)
                node_ref["description"] = json.dumps(
                    data_ref, sort_keys=True, indent=4, separators=(",", ": ")
                )

    def _add_link(self, link: dict, link_data: dict = {}) -> None:
        link_hash = self._make_hash_tuple(link)
        self.links_dict.setdefault(link_hash, [])
        if link not in self.links_dict[link_hash]:
            if link_data and self.add_data:
                link["description"] = json.dumps(
                    link_data, sort_keys=True, indent=4, separators=(",", ": ")
                )
            self.links_dict[link_hash].append(link)

    def _pack_links(self) -> None:
        """
        Method to iterate over links between node pairs and pack links based
        on the local and peer interface IDs
        """
        for hash in self.links_dict.keys():
            links = self.links_dict[hash]
            # continue if only one link between node pairs
            if len(links) <= 1:
                continue
            self.links_dict[hash] = []
            while links:
                link = links.pop()
                pair_link_index = None
                local_intf_id = link.pop("local_intf_id")
                peer_intf_id = link.pop("peer_intf_id")
                peer_ip = link.pop("peer_ip")
                local_ip = link.pop("local_ip")
                # try to find link pair using ISIS interface ID
                if local_intf_id and peer_intf_id:
                    for link_2 in links:
                        if (
                            link_2["peer_intf_id"] == local_intf_id
                            and link_2["local_intf_id"] == peer_intf_id
                        ):
                            link["trgt_label"] = link_2["src_label"]
                            self._add_link(
                                link=link,
                                link_data={
                                    **json.loads(link["description"]),
                                    **json.loads(link_2["description"]),
                                },
                            )
                            # form new link label if PID does not match
                            if link["label"] != link_2["label"]:
                                pid_1, level_1 = link["label"].split(":")
                                pid_2, level_2 = link_2["label"].split(":")
                                link["label"] = "{}:{}".format(
                                    pid_1
                                    if pid_1 == pid_2
                                    else "{}-{}".format(pid_1, pid_2),
                                    level_1,
                                )
                            pair_link_index = links.index(link_2)
                            break
                # try to find link pair using interface IP addresses
                elif local_ip and peer_ip:
                    for link_2 in links:
                        if (
                            link_2["peer_ip"] == local_ip
                            and link_2["local_ip"] == peer_ip
                        ):
                            link["trgt_label"] = link_2["src_label"]
                            self._add_link(
                                link=link,
                                link_data={
                                    **json.loads(link["description"]),
                                    **json.loads(link_2["description"]),
                                },
                            )
                            # form new link label if PID does not match
                            if link["label"] != link_2["label"]:
                                pid_1, level_1 = link["label"].split(":")
                                pid_2, level_2 = link_2["label"].split(":")
                                link["label"] = "{}:{}".format(
                                    pid_1
                                    if pid_1 == pid_2
                                    else "{}-{}".format(pid_1, pid_2),
                                    level_1,
                                )
                            pair_link_index = links.index(link_2)
                            break
                # remove link from links if it was combined
                if pair_link_index is not None:
                    _ = links.pop(pair_link_index)
                # add link back to links if have not found a match for it
                else:
                    self._add_link(link)

    def _lookup_rid(self):
        """
        Method to lookup RID in lookup data and update node labels.
        """
        for node in self.nodes_dict.values():
            node_ip = node["top_label"]
            if node_ip in self.ip_lookup_data:
                node.update(
                    {
                        k: v
                        for k, v in self.ip_lookup_data[node_ip].items()
                        if k in ["top_label", "bottom_label", "label"]
                    }
                )

    def _lookup_ip_interfaces(self):
        """
        Method to search for link IP addresses in lookup table and add
        interface names to the link labels.
        """
        for links in self.links_dict.values():
            for link in links:
                # modify source label
                if link.get("src_label"):
                    link_src_ip, link_src_metric = link["src_label"].split(":")
                    # skip link_src_ip if its not IPv4 address
                    if link_src_ip.count(".") != 3:
                        continue
                    if self.ip_lookup_data.get(link_src_ip, {}).get("interface"):
                        link["src_label"] = "{}:{}:{}".format(
                            self.ip_lookup_data[link_src_ip]["interface"],
                            link_src_ip,
                            link_src_metric,
                        )
                # modify target label
                if link.get("trgt_label"):
                    link_trgt_ip, link_trgt_metric = link["trgt_label"].split(":")
                    # skip link_trgt_ip if its not IPv4 address
                    if link_trgt_ip.count(".") != 3:
                        continue
                    if self.ip_lookup_data.get(link_trgt_ip, {}).get("interface"):
                        link["trgt_label"] = "{}:{}:{}".format(
                            self.ip_lookup_data[link_trgt_ip]["interface"],
                            link_trgt_ip,
                            link_trgt_metric,
                        )

    def _update_drawing(self):
        """
        Method to add formed links and nodes to the drawing object
        """
        self.graph_dict["nodes"] = list(self.nodes_dict.values())
        for i in self.links_dict.values():
            self.graph_dict["links"].extend(i)
        # import pprint; pprint.pprint(self.graph_dict, width =100)
        self.drawing.from_dict(self.graph_dict)



"""
CLI OSPFv2 LSDB Data Plugin
***************************

CLI OSPFv2 LSDB Data Plugin can process network devices CLI output of OSPFv2 LSDB content to
populate N2G drawing with OSPF topology nodes and links.

CLI output from devices parsed using TTP Templates into a dictionary structure.

After parsing, results processed further to form a dictionary of nodes and links keyed
by unique nodes and links identifiers wit values being nodes dictionaries and for links
it is a list of dictionaries of links between same pair of nodes. For nodes OSPF RID
used as a unique ID, for links it is sorted tuple of ``source``, ``target`` and ``label``
keys' values. This structure helps to eliminate duplicates.

Next step is post processing, such as packing links between nodes. By default cli_ospf_data
tries to match and pack nodes based on the IP addresses and their subnets, it checks
if IP addresses are part of same subnet using prefix lengths - 31, 30, 29, ... 22 - if
IP addresses happens to be part of same subnet, link packed in one link.

Last step is to populate N2G drawing with new nodes and links using ``from_dict`` method.

Features Supported
------------------

**Support matrix**

+---------------+------------+------------+------------+------------+-----------+-----------+
| Platform      | Router     | OSPF       | External   | Summary    | interface | interface |
| Name          | LSA        | Peers      | LSA        | LSA        | config    | state     |
+===============+============+============+============+============+===========+===========+
| cisco_ios     |     YES    |     ---    |     ---    |     ---    |    ---    |    ---    |
+---------------+------------+------------+------------+------------+-----------+-----------+
| cisco_xr      |     YES    |     ---    |     ---    |     ---    |    ---    |    ---    |
+---------------+------------+------------+------------+------------+-----------+-----------+
| cisco_nxos    |     ---    |     ---    |     ---    |     ---    |    ---    |    ---    |
+---------------+------------+------------+------------+------------+-----------+-----------+
| huawei        |     YES    |     ---    |     ---    |     ---    |    ---    |    ---    |
+---------------+------------+------------+------------+------------+-----------+-----------+

Required Commands output
------------------------

cisco_ios:

* ``show ip ospf database router`` - mandatory, used to source nodes and links for topology
* ``show ip ospf database summary``
* ``show ip ospf database external``

cisco_xr:

* ``show ospf database router`` - mandatory, used to source nodes and links for topology
* ``show ospf database summary``
* ``show ospf database external``

huawei:

* ``display ospf lsdb router`` - mandatory, used to source nodes and links for topology

Sample usage
------------

Code to populate yEd diagram object with OSPF LSDB sourced nodes and links::

    from N2G import cli_l2_data, yed_diagram

    data = {"cisco_xr": ['''
    RP/0/RP0/CPU0:router-1#show ospf database router

                OSPF Router with ID (10.0.1.1) (Process ID 1)

                    Router Link States (Area 0.0.0.0)

      LS age: 406
      Options: (No TOS-capability, DC)
      LS Type: Router Links
      Link State ID: 10.0.1.1
      Advertising Router: 10.0.1.1
      LS Seq Number: 8000010c
      Checksum: 0x24dd
      Length: 132
       Number of Links: 9

        Link connected to: another Router (point-to-point)
         (Link ID) Neighboring Router ID: 10.0.1.4
         (Link Data) Router Interface address: 0.0.0.12
          Number of TOS metrics: 0
           TOS 0 Metrics: 1100

        Link connected to: another Router (point-to-point)
         (Link ID) Neighboring Router ID: 10.0.1.2
         (Link Data) Router Interface address: 0.0.0.10
          Number of TOS metrics: 0
           TOS 0 Metrics: 1100

      Routing Bit Set on this LSA
      LS age: 1604
      Options: (No TOS-capability, DC)
      LS Type: Router Links
      Link State ID: 10.0.1.2
      Advertising Router: 10.0.1.2
      LS Seq Number: 8000010b
      Checksum: 0xdc96
      Length: 132
       Number of Links: 9

        Link connected to: another Router (point-to-point)
         (Link ID) Neighboring Router ID: 10.0.1.3
         (Link Data) Router Interface address: 0.0.0.52
          Number of TOS metrics: 0
           TOS 0 Metrics: 1100

        Link connected to: another Router (point-to-point)
         (Link ID) Neighboring Router ID: 10.0.1.4
         (Link Data) Router Interface address: 0.0.0.53
          Number of TOS metrics: 0
           TOS 0 Metrics: 1100
        ''']
    }

    drawing = yed_diagram()
    drawer = cli_ospf_data(drawing)
    drawer.work(data)
    drawer.drawing.dump_file()

API Reference
-------------

.. autoclass:: N2G.plugins.data.cli_ospf_data.cli_ospf_data
   :members:
"""
import logging
import json
import os
import csv
import ipaddress
from fnmatch import fnmatchcase

try:
    from ttp import ttp
    from ttp_templates import get_template

    HAS_TTP = True
except ImportError:
    HAS_TTP = False

# initiate logging
log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# Main class:
# -----------------------------------------------------------------------------


class cli_ospf_data:
    """
    Main class to instantiate OSPFv2 LSDB CLI Data Plugin object.

    :param drawing: (obj) N2G Diagram object
    :param ttp_vars: (dict) Dictionary to use as vars attribute while instantiating TTP parser object
    :param ip_lookup_data: (dict or str) IP Lookup dictionary or OS path to CSV file
    :param add_connected: (bool) if True, will add connected subnets as nodes, default is False
    :param ptp_filter: (list) list of glob patterns to filter point-to-point links based on link IP
    :param add_data: (bool) if True (default) adds data information to nodes and links

    ``ip_lookup_data`` dictionary must be keyed by OSPF RID IP address, with values
    being dictionary which must contain ``hostname`` key with optional additional keys
    to use for N2G diagram module node, e.g. ``label``, ``top_label``, ``bottom_label``,
    ``interface``etc. If ``ip_lookup_data`` is an OS path to CSV file, that file's first
    column header must be ``ip`` , file must contain ``hostname`` column, other columns
    values set to N2G diagram module node attributes, e.g. ``label``, ``top_label``,
    ``bottom_label``, ``interface`` etc.

    If lookup data contains ``interface`` key, it will be added to link label.

    Sample ip_lookup_data dictionary::

        {
            "1.1.1.1": {
                "hostname": "router-1",
                "bottom_label": "1 St address, City X",
                "interface": "Gi1"
            }
        }

    Sample ip_lookup_data CSV file::

        ip,hostname,bottom_label,interface
        1.1.1.1,router-1,"1 St address, City X",Gi1

    ``ptp_filter`` default list of patterns are:

    * ``0*`` - Cisco MPLS TE forwarding adjacencies links
    * ``112*`` - huawei DCN OSPF links
    """

    def __init__(
        self,
        drawing,
        ttp_vars: dict = None,
        ip_lookup_data: dict = None,
        add_connected: bool = False,
        ptp_filter: list = None,
        add_data: bool = True,
    ):
        self.ttp_vars = ttp_vars or {}
        self.drawing = drawing
        self.drawing.node_duplicates = "update"
        self.add_connected = add_connected
        self.ptp_filter = ptp_filter or ["0*", "112*"]
        self.add_data = add_data
        self.parsed_data = {}
        self.nodes_dict = {}
        self.links_dict = {}
        self.graph_dict = {"nodes": [], "links": []}
        self.ip_lookup_data = ip_lookup_data or {}
        self._load_ip_lookup_data()

    def _load_ip_lookup_data(self) -> None:
        """
        Helper function to load CSV table in a dictionary keyed
        by values in first column. This dictionary further used
        to perform IP to node details lookup for OSPF router ID.
        """
        # load lookup data
        if self.ip_lookup_data and isinstance(self.ip_lookup_data, str):
            with open(self.ip_lookup_data) as f:
                reader = csv.DictReader(f)
                self.ip_lookup_data = {r["ip"]: r for r in reader if "ip" in r}

    def work(self, data):
        """
        Method to parse OSPF LSDB data and add nodes and links to N2G drawing.

        :param data: (dict or str) dictionary keyed by platform name or OS path
            string to directories with text files

        If data is dictionary, keys must correspond to "Platform" column in
        *Supported platforms* table, values are lists of text items to
        process.

        Data dictionary sample::

            data = {
                "cisco_ios" : ["h1", "h2"],
                "cisco_ios": ["h3", "h4"],
                "cisco_nxos": ["h5", "h6"],
                ...etc...
            }

        Where ``hX`` device's show commands output.

        If data is an OS path directory string, child directories' names must correspond
        to **Platform** column in `Features Supported`_ section table. Each child directory
        should contain text files with show commands output for each device, names of files
        are arbitrary, but output should contain device prompt to extract device hostname.

        Directories structure sample::

            ├───folder_with_data
                ├───cisco_ios
                │       switch1.txt
                │       switch2.txt
                └───cisco_nxos
                        nxos_switch_1.txt
                        nxos_switch_2.txt

        To point N2G to above location ``data`` attribute string can be ``/var/data/n2g/folder_with_data/``
        """
        self._parse(data)
        self._form_base_graph_dict()
        self._pack_links()
        if self.ip_lookup_data:
            self._lookup_rid()
            self._lookup_ip_interfaces()
        self._update_drawing()

    def _make_hash_tuple(self, data: dict) -> tuple:
        """
        Helper function to form Edge tuple to use as a hash to
        identify all the links between same pair of nodes using
        ``source``, ``target`` and ``label`` keys' values.

        :param data: (dict) link dictionary with source, target and label keys
        """
        return tuple(sorted([data["source"], data["target"], data.get("label", "")]))

    def _parse(self, data: [dict, str]) -> None:
        """
        Method to parse data using TTP Templates

        :param data: (dict or str) dictionary of data items or OS
            path to folders with data to parse
        :return: None
        """
        if not HAS_TTP:
            raise ModuleNotFoundError(
                "N2G:cli_ospf_data failed importing TTP, is it installed?"
            )
        parser = ttp(vars=self.ttp_vars, log_level="ERROR")
        # process data dictionary
        if isinstance(data, dict):
            for platform_name, text_list in data.items():
                ttp_template = get_template(
                    misc="N2G/cli_ospf_data/{}.txt".format(platform_name)
                )
                parser.add_template(template=ttp_template, template_name=platform_name)
                for item in text_list:
                    parser.add_input(item, template_name=platform_name)
        # process directories at OS path
        elif isinstance(data, str):
            # get all sub-folders and load respective templates
            with os.scandir(data) as dirs:
                for entry in dirs:
                    if entry.is_dir():
                        platform_name = entry.name
                        ttp_template = get_template(
                            misc="N2G/cli_ospf_data/{}.txt".format(platform_name)
                        )
                        parser.add_template(
                            template=ttp_template, template_name=platform_name
                        )
                        parser.add_input(
                            data=os.path.abspath(entry), template_name=platform_name
                        )
        else:
            raise TypeError(
                "Expecting dictionary or string, but '{}' given".format(type(data))
            )
        parser.parse(one=True)
        self.parsed_data = parser.result(structure="flat_list")
        # import pprint; pprint.pprint(self.parsed_data, width = 100)

    def _process_router_lsa(
        self, router_lsa: dict, ospf_pid: str, ospf_data: dict, device: dict
    ) -> None:
        """"""
        # make node out of router LSA
        self._add_node(
            node={
                "id": router_lsa["originator_rid"],
                "label": router_lsa["originator_rid"],
                "bottom_label": "Node",
            },
            node_data={
                "lsdb_source": {
                    "hostname": self.ip_lookup_data.get(ospf_data["local_rid"], {}).get(
                        "hostname"
                    ),
                    "ospf_rid": ospf_data["local_rid"],
                    "ospf_pid": ospf_pid,
                },
                **router_lsa,
            },
        )
        # go over PTP peers
        for ptp_peer in router_lsa.get("ptp_peers", []):
            # ignore OSPF adjacencies
            if any([fnmatchcase(ptp_peer["link_data"], p) for p in self.ptp_filter]):
                continue
            self._add_link(
                link={
                    "source": router_lsa["originator_rid"],
                    "src_label": "{}:{}".format(
                        ptp_peer["link_data"], ptp_peer["metric"]
                    ),
                    "label": "A:{}".format(router_lsa["area"]),
                    "target": ptp_peer["link_id"],
                    "source_ip": ptp_peer["link_data"],
                }
            )
        # go over BMA peers
        for bma_peer in router_lsa.get("bma_peers", []):
            # find BMA subnet
            bma_ip = ipaddress.IPv4Address(bma_peer["link_id"])
            bma_subnet = ""
            for connet in router_lsa.get("connected_stub", []):
                net = ipaddress.IPv4Network(
                    "{}/{}".format(connet["link_id"], connet["link_data"])
                )
                if bma_ip in net:
                    bma_subnet = str(net)
                    break
            # add BMA DR Node
            self._add_node(
                node={
                    "id": "DR {}".format(bma_peer["link_id"]),
                    "label": "DR {}".format(bma_peer["link_id"]),
                    "bottom_label": "DR Node",
                    "top_label": bma_subnet,
                }
            )
            self._add_link(
                link={
                    "source": router_lsa["originator_rid"],
                    "src_label": "{}:{}".format(
                        bma_peer["link_data"], bma_peer["metric"]
                    ),
                    "label": "A:{}".format(router_lsa["area"]),
                    "target": "DR {}".format(bma_peer["link_id"]),
                }
            )
        # go over connected subnets
        if self.add_connected:
            # form a list of IP address for router ptp links and BMA subnets
            node_peering_ip = [
                ipaddress.IPv4Address(i["link_data"])
                for i in router_lsa.get("ptp_peers", [])
                if not i["link_data"].startswith("0.")
            ]
            node_peering_ip.extend(
                [
                    ipaddress.IPv4Address(i["link_id"])
                    for i in router_lsa.get("bma_peers", [])
                ]
            )
            for connet in router_lsa.get("connected_stub", []):
                net = ipaddress.IPv4Network(
                    "{}/{}".format(connet["link_id"], connet["link_data"])
                )
                # filter networks that already formed ptp links
                for ip in node_peering_ip:
                    if ip in net:
                        break
                else:
                    # add subnet node and links if its not used to form ptp or BMA peering
                    self._add_node(
                        node={
                            "id": str(net),
                            "label": str(net),
                            "bottom_label": "Subnet",
                        }
                    )
                    self._add_link(
                        link={
                            "source": router_lsa["originator_rid"],
                            "src_label": "M:{}".format(connet["metric"]),
                            "label": "A:{}".format(router_lsa["area"]),
                            "target": str(net),
                        }
                    )

    def _process_external_lsa(
        self, external_lsa: dict, ospf_pid: str, ospf_data: dict, device: dict
    ) -> None:
        pass

    def _process_summary_lsa(
        self, summry_lsa: dict, ospf_pid: str, ospf_data: dict, device: dict
    ) -> None:
        pass

    def _form_base_graph_dict(self) -> None:
        for device in self.parsed_data:
            # go over all OSPF processes on the box
            for ospf_pid, ospf_data in device.get("ospf_processes", {}).items():
                # process LSA
                for router_lsa in ospf_data.get("router_lsa", []):
                    self._process_router_lsa(router_lsa, ospf_pid, ospf_data, device)
                for external_lsa in ospf_data.get("external_lsa", []):
                    self._process_external_lsa(
                        external_lsa, ospf_pid, ospf_data, device
                    )
                for summry_lsa in ospf_data.get("summry_lsa", []):
                    self._process_summary_lsa(summry_lsa, ospf_pid, ospf_data, device)

    def _add_node(self, node: dict, node_data: dict = None) -> None:
        node_data = node_data or {}
        # add new node
        if not node["id"] in self.nodes_dict:
            if node_data and self.add_data:
                node["description"] = json.dumps(
                    node_data, sort_keys=True, indent=4, separators=(",", ": ")
                )
            self.nodes_dict[node["id"]] = node
        # update node attributes if they do not exists already
        else:
            stored_node = self.nodes_dict[node["id"]]
            for key, value in node.items():
                if not key in stored_node:
                    stored_node[key] = value
            if not "description" in stored_node and node_data and self.add_data:
                stored_node["description"] = json.dumps(
                    node_data, sort_keys=True, indent=4, separators=(",", ": ")
                )

    def _add_link(self, link: dict, link_data: dict = None) -> None:
        link_data = link_data or {}
        link_hash = self._make_hash_tuple(link)
        self.links_dict.setdefault(link_hash, [])
        if link not in self.links_dict[link_hash]:
            if link_data and self.add_data:
                link["description"] = json.dumps(
                    link_data, sort_keys=True, indent=4, separators=(",", ": ")
                )
            self.links_dict[link_hash].append(link)

    def _pack_links(self) -> None:
        """
        Method to iterate over links between node pairs and pack links based
        on these criteria:
          - combine links for /30 or /29 subnets
        """
        for hash in self.links_dict.keys():
            links = self.links_dict[hash]
            # continue if only one link between node pairs
            if len(links) <= 1:
                continue
            self.links_dict[hash] = []
            while links:
                link = links.pop()
                pair_link_index = None
                link_ip_addr = link.pop("source_ip")
                # try to match on subnets
                for m in [31, 30, 29, 28, 27, 26, 25, 24, 23, 22]:
                    link_ip = ipaddress.IPv4Interface("{}/{}".format(link_ip_addr, m))
                    for link_2 in links:
                        link_2_ip = ipaddress.IPv4Interface(
                            "{}/30".format(link_2["source_ip"])
                        )
                        if link_2_ip in link_ip.network:
                            link["trgt_label"] = link_2["src_label"]
                            self._add_link(
                                link=link,
                                link_data={
                                    "subnet": str(link_ip.network),
                                    "ospf_area": link["label"].split(":")[1],
                                    "source_rid": link["source"],
                                    "target_rid": link["target"],
                                    "source_ip": str(link_ip.ip),
                                    "target_ip": str(link_2_ip.ip),
                                },
                            )
                            pair_link_index = links.index(link_2)
                            break
                    if pair_link_index:
                        break
                # remove link from links if it was combined
                if pair_link_index is not None:
                    _ = links.pop(pair_link_index)
                # add link back to links if have not found a match for it
                else:
                    self._add_link(link)

    def _lookup_rid(self):
        """
        Method to lookup RID in lookup data and use hostname as label and ID,
        this method also updates all the links with hostnames as source and target
        node IDs
        """
        for node in self.nodes_dict.values():
            if node["id"] in self.ip_lookup_data:
                node_id = node["id"]
                node["id"] = self.ip_lookup_data[node_id].get("hostname", node_id)
                node["label"] = self.ip_lookup_data[node_id].get(
                    "hostname", node["label"]
                )
                node.update(
                    {
                        k: v
                        for k, v in self.ip_lookup_data[node_id].items()
                        if k != "interface"
                    }
                )
        for links in self.links_dict.values():
            for link in links:
                if link["source"] in self.ip_lookup_data:
                    link["source"] = self.ip_lookup_data[link["source"]].get(
                        "hostname", link["source"]
                    )
                if link["target"] in self.ip_lookup_data:
                    link["target"] = self.ip_lookup_data[link["target"]].get(
                        "hostname", link["target"]
                    )

    def _lookup_ip_interfaces(self):
        """
        Method to search for link IP addresses in lookup table and add
        interface details to the link labels.
        """
        for links in self.links_dict.values():
            for link in links:
                # modify source label
                if link.get("src_label"):
                    link_src_ip, link_src_metric = link["src_label"].split(":")
                    if self.ip_lookup_data.get(link_src_ip, {}).get("interface"):
                        link["src_label"] = "{}:{}:{}".format(
                            self.ip_lookup_data[link_src_ip]["interface"],
                            link_src_ip,
                            link_src_metric,
                        )
                # modify target label
                if link.get("trgt_label"):
                    link_trgt_ip, link_trgt_metric = link["trgt_label"].split(":")
                    if self.ip_lookup_data.get(link_trgt_ip, {}).get("interface"):
                        link["trgt_label"] = "{}:{}:{}".format(
                            self.ip_lookup_data[link_trgt_ip]["interface"],
                            link_trgt_ip,
                            link_trgt_metric,
                        )

    def _update_drawing(self):
        """
        Method to add formed links and nodes to the drawing object
        """
        self.graph_dict["nodes"] = list(self.nodes_dict.values())
        for i in self.links_dict.values():
            self.graph_dict["links"].extend(i)
        # import pprint; pprint.pprint(self.graph_dict, width =100)
        self.drawing.from_dict(self.graph_dict)



"""
yEd SVG Viewer
**************

yED SVG Viewer allows to start simple Flask WEB UI application to visualize network
data using `D3.js <https://d3js.org/>`_ library.

This viewer needs to have Flask installed::

    pip install flask

Flask installed as part of ``full`` extras as well.

To run using N2G CLI tool::

    N2G --yed-svg-viewer --diagrams-dir /Diagrams/

Access WEB UI application via URL ``http://127.0.0.1:9000`` using your browser.

By default Flask server starts and listens on all operating system interfaces, but
specific IP address and port number can be specified as required using ``--ip`` and
``--port`` N2G CLI arguments.

OS path to directory with diagram files can be specified using ``--diagrams-dir``
N2G CLI tool argument or using ``N2G_DIAGRAMS_DIR`` environment variable. If no
``--diagrams-dir`` argument provided, N2G attempts to retrieve diagrams directory
path using ``N2G_DIAGRAMS_DIR`` environment variable.

Diagrams directory should contain SVG files produced by yED Graph Editor application
using ``File -> Export -> Save as type: SVG format`` feature.
"""
import os
import json
import logging

try:
    from flask import Flask, render_template

    HAS_FLASK = True
except ImportError:
    HAS_FLASK = False

log = logging.getLogger(__name__)

app_dir = os.path.abspath(os.path.dirname(__file__))

if HAS_FLASK:
    app = Flask(
        __name__, static_folder=app_dir, static_url_path="", template_folder=app_dir
    )

    @app.route("/")
    def home():
        return render_template("yed_viewer.html", select2_menu_data=scan_graphs())

    @app.route("/diagrams/<diagram_name>")
    def get_diagram(diagram_name):
        with open(os.path.join(diagrams_directory, diagram_name)) as f:
            return f.read()


def scan_graphs() -> str:
    """
    Helper function to scan folder with diagrams. Return
    json string of list of diagram dictionaries.
    """
    select2_menu_data = [
        {"id": i, "text": i.split(".")[0]}
        for i in os.listdir(diagrams_directory)
        if i.endswith("svg")
    ]
    return json.dumps(select2_menu_data)


def run_yed_viewer(
    ip: str = "0.0.0.0",
    port: int = 9000,
    debug: bool = True,
    diagrams_dir: str = None,
    **kwargs,
) -> None:
    """
    Function to start Flask server to run yED viewer application.

    :param ip: IP address to run server on
    :param port: port number to run server on
    :param debug: whether server should run in debug mode
    :param diagrams_dir: OS path to directory with diagrams, by default diagrams directory
        sourced using ``N2G_DIAGRAMS_DIR`` environment variable
    :param kwargs: any additional argument to pass to Flask app.run call
    """
    global diagrams_directory
    diagrams_directory = diagrams_dir or os.getenv("N2G_DIAGRAMS_DIR", "diagrams")
    log.debug(
        f"Starting server on 'http://{ip}:{port}', diagrams directory: '{diagrams_directory}'"
    )
    app.run(host=ip, port=port, debug=debug, **kwargs)


if __name__ == "__main__":
    run_yed_viewer()



import hashlib
import os
import logging
import json

# initiate logging
log = logging.getLogger(__name__)


class v3d_diagramm:
    """
    Class to produce JSON data structure compatible with 
    `3D Force-Directed Graph <https://github.com/vasturiano/3d-force-graph>`_ 
    library `JSON input syntax <https://github.com/vasturiano/3d-force-graph#input-json-syntax>`_
    
    :param node_duplicates: (str) what to do with node duplicates - ``skip`` (default), ``update`` or ``log``
    :param link_duplicates: (str) what to do with link duplicates - ``skip`` (default), ``update`` or ``log``
    """

    def __init__(self, node_duplicates="skip", link_duplicates="skip"):
        self.drawing = {"nodes": [], "links": []}
        self.node_duplicates = node_duplicates
        self.link_duplicates = link_duplicates
        self.nodes_dict = {}
        self.links_dict = {}

    def _node_exists(self, id, **kwargs):
        # check if node with given id already exists
        if id in self.nodes_dict:
            if self.node_duplicates == "log":
                log.error("add_node: node '{}' already added to graph".format(id))
            elif self.node_duplicates == "skip":
                pass
            elif self.node_duplicates == "update":
                self.update_node(id=id, **kwargs)
            return True
        else:
            return False

    def _link_exists(self, id, **kwargs):
        """method, used to check duplicate edges"""
        # check if edge with given id already exists
        if id in self.links_dict:
            if self.link_duplicates == "log":
                log.error(
                    "_link_exists: '{}' edge already exists, data {}".format(id, kwargs)
                )
            elif self.link_duplicates == "skip":
                pass
            elif self.link_duplicates == "update":
                self.update_link(id=id, **kwargs)
            return True
        else:
            return False

    def add_node(
        self, id, label="", data=None, color="green", nodeResolution=8, **kwargs
    ):
        """
        Method to add node to the diagram.

        :param id: (str) mandatory, unique node identifier, usually equal to node name
        :param label: (str) node label, if not provided, set equal to id
        :param data: (dict) dictionary of key value pairs to add as node data
        :param fx: (int) node position on x axis
        :param fy: (int) node position on y axis
        :param fz: (int) node position on z axis
        :param color: (str) node color e.g. ``blue``, default is ``green``
        :param nodeResolution: (int) geometric resolution of the node, expressed in how 
          many slice segments to divide the circumference. Higher values yield smoother spheres.
        :param kwargs: (dict) any additional kwargs to add to node dictionary as per
          `node styling attributes <https://github.com/vasturiano/3d-force-graph#node-styling>`_
          such as ``nodeRelSize``, ``nodeOpacity``, ``nodeVal`` etc.
        """
        data = data or {}
        # process node
        if self._node_exists(id, label=label, data=data):
            return
        if not label.strip():
            label = id
        # create node element
        node = {
            "id": id,
            "label": label,
            "color": color,
            "nodeResolution": nodeResolution,
            "data": data,
            **kwargs,
        }
        # add node to nodes dictionary
        self.nodes_dict[id] = node

    def update_node(self, id, data=None, **kwargs):
        """
        Method to update node details. Uses node ``id`` to search for node to update

        :param id: (str) mandatory, unique node identifier
        :param data: (dict) data argument/key dictionary content to update existing values
        :param kwargs: (dict) any additional arguments to update node dictionary
        """
        data = data or {}
        node = self.nodes_dict.get(id)
        if not node:
            log.error("node_update: node {} not found".format(id))
            return

        # update node attributes
        node["data"].update(data)
        node.update(kwargs)

    def delete_node(self, id):
        """
        Method to delete node. Uses node ``id`` to search for node to delete.

        :param id: (str) mandatory, unique node identifier
        """
        _ = self.nodes_dict.pop(id, None)
        # delete all bound links
        for lid in list(self.links_dict.keys()):
            if (
                self.links_dict[lid]["source"] == id
                or self.links_dict[lid]["target"] == id
            ):
                _ = self.links_dict.pop(lid, None)

    def _make_edge_id(self, source, target, label="", src_label="", trgt_label=""):
        edge_tup = tuple(sorted([label, source, target, src_label, trgt_label]))
        return hashlib.md5(",".join(edge_tup).encode()).hexdigest()

    def add_link(
        self,
        source,
        target,
        label="",
        src_label="",
        trgt_label="",
        data=None,
        id=None,
        **kwargs
    ):
        """
        Method to add link between nodes.

        :param source: (str) mandatory, source node id
        :param target: (str) mandatory, target node id
        :param label: (str) link label to display at the center of the link
        :param data: (dict) dictionary of key value pairs to add as link data
        :param src_label: (str) link label to use next to source node
        :param trgt_label: (str) link label to use next to target node
        :param id: (str) explicit link identifier to use
        :param kwargs: (dict) any additional kwargs to add to link dictionary
        
        .. note:: If source or target nodes does not exists, they will be automatically
          created

        All labels are optional and substituted with empty values to calculate link id.
        
        By default V3D uses below code to produce MD5 hash digest for link id::
        
            link_tup = tuple(sorted([label, source, target, src_label, trgt_label]))
            link_id = hashlib.md5(",".join(edge_tup).encode()).hexdigest()
        """
        data = data or {}
        # check type of source and target attribute
        source_node_dict = source.copy() if isinstance(source, dict) else {"id": source}
        source = source_node_dict.pop("id")
        target_node_dict = target.copy() if isinstance(target, dict) else {"id": target}
        target = target_node_dict.pop("id")
        # check if target and source nodes exist, add it if not, self._node_exists method will update node
        # if self.node_duplicates set to update, by default its set to skip
        if not self._node_exists(source, **source_node_dict):
            self.add_node(id=source, **source_node_dict)
        if not self._node_exists(target, **target_node_dict):
            self.add_node(id=target, **target_node_dict)
        # create edge id
        edge_id = id or self._make_edge_id(source, target, label, src_label, trgt_label)
        if self._link_exists(
            id=edge_id,
            source=source,
            target=target,
            new_label=label,
            new_src_label=src_label,
            new_trgt_label=trgt_label,
            data=data,
            **kwargs
        ):
            return
        # create link
        link = {
            "id": edge_id,
            "label": label,
            "source": source,
            "target": target,
            "src_label": src_label,
            "trgt_label": trgt_label,
            "data": data,
            **kwargs,
        }
        # save link to graph
        self.links_dict[edge_id] = link

    def update_link(
        self,
        source=None,
        target=None,
        label="",
        src_label="",
        trgt_label="",
        new_label="",
        new_src_label="",
        new_trgt_label="",
        data=None,
        url="",
        id=None,
        **kwargs
    ):
        """
        Method to update link details. Uses link ``id`` to search for link to update, if no ``id`` 
        provided uses ``source, target, label, src_label, trgt_label`` to calculate edge id.
        
        :param source: (str) source node id
        :param target: (str) target node id
        :param label: (str) existing link label
        :param src_label: (str) existing link source label
        :param trgt_label: (str) existing link target label
        :param new_label: (str) new link label to replace existing label
        :param new_src_label: (str) new link source label to replace existing source label
        :param new_trgt_label: (str) new link target label to replace existing target label
        :param data: (dict) dictionary of key value pairs to update link data
        :param url: (str) url string to save as link ``url`` attribute
        :param id: (str) link identifier to find the link to update
        :param kwargs: (dict) any additional kwargs to update link dictionary
        """
        data = data or {}
        edge_id = id or self._make_edge_id(source, target, label, src_label, trgt_label)
        link = self.links_dict.get(edge_id)

        if not link:
            log.error("link_update: link {} not found".format(edge_id))
            return

        # update link attributes
        if data:
            link["data"].update(data)
        if url:
            link["url"] = url

        # update labels and edge_id
        if any([new_label, new_src_label, new_trgt_label]):
            link["label"] = new_label if new_label else link["label"]
            link["src_label"] = new_src_label if new_src_label else link["src_label"]
            link["trgt_label"] = (
                new_trgt_label if new_trgt_label else link["trgt_label"]
            )
            link["id"] = self._make_edge_id(
                source, target, link["label"], link["src_label"], link["trgt_label"]
            )

        # update remaining attributes
        link.update(kwargs)

    def delete_link(
        self, source=None, target=None, label="", src_label="", trgt_label="", id=None
    ):
        """
        Method to delete link. Uses link ``id`` to search for link to delete, if no ``id`` 
        provided uses ``source, target, label, src_label, trgt_label`` to calculate edge id.
        
        :param source: (str) source node id
        :param target: (str) target node id
        :param label: (str) existing link label
        :param src_label: (str) link source label
        :param trgt_label: (str) link target label
        :param id: (str) link identifier to find the link to delete
        """
        edge_id = id or self._make_edge_id(source, target, label, src_label, trgt_label)
        _ = self.links_dict.pop(edge_id, None)

    def layout(self, algo="kk3d", dx=100, dy=100, dz=100, **kwargs):
        """
        Method to calculate graph layout using Python
        `igraph <https://igraph.org/python/doc/tutorial/tutorial.html#layout-algorithms>`_
        library

        :param algo: (str) name of igraph layout algorithm to use, default is 'kk3d'. Reference
          `Layout algorithms` table below for valid algo names
        :param kwargs: (dict) any additional kwargs to pass to igraph ``Graph.layout`` method

        **Layout algorithms**

        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | algo name                       |    description                                                                                                 |
        +=================================+================================================================================================================+
        | circle, circular                | Deterministic layout that places the vertices on a circle                                                      |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | drl                             | The Distributed Recursive Layout algorithm for large graphs                                                    |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | fr                              | Fruchterman-Reingold force-directed algorithm                                                                  |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | fr3d, fr_3d                     | Fruchterman-Reingold force-directed algorithm in three dimensions                                              |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | grid_fr                         | Fruchterman-Reingold force-directed algorithm with grid heuristics for large graphs                            |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | kk                              | Kamada-Kawai force-directed algorithm                                                                          |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | kk3d, kk_3d                     | Kamada-Kawai force-directed algorithm in three dimensions                                                      |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | large, lgl, large_graph         | The Large Graph Layout algorithm for large graphs                                                              |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | random                          | Places the vertices completely randomly                                                                        |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | random_3d                       | Places the vertices completely randomly in 3D                                                                  |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | rt, tree                        | Reingold-Tilford tree layout, useful for (almost) tree-like graphs                                             |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | rt_circular, tree               | Reingold-Tilford tree layout with a polar coordinate post-transformation, useful for (almost) tree-like graphs |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | sphere, spherical, circular_3d  | Deterministic layout that places the vertices evenly on the surface of a sphere                                |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        
        .. note:: if 2d layout algorithm called, z axis coordinate set to 0
        """
        try:
            from igraph import Graph as ig
        except ImportError:
            raise SystemExit(
                "Failed to import igraph, install - pip install python-igraph"
            )
        igraph_graph = ig()
        # iterate over nodes
        for item in self.nodes_dict.values():
            igraph_graph.add_vertex(name=item["id"])
        # add edges
        for item in self.links_dict.values():
            igraph_graph.add_edge(source=item["source"], target=item["target"])
        # calculate layout
        layout = igraph_graph.layout(layout=algo, **kwargs)
        # scale layout to diagram size
        if "3d" in algo.lower().strip():
            layout.fit_into(bbox=(dx, dy, dz))
        else:
            layout.fit_into(bbox=(dx, dy))
        # add coordinates from layout to nodes
        for index, coord_item in enumerate(layout.coords):
            if "3d" in algo.lower().strip():
                x_coord, y_coord, z_coord = coord_item
            else:
                x_coord, y_coord, z_coord = (*coord_item, 0)
            node_id = igraph_graph.vs[index].attributes()["name"]
            node_element = self.nodes_dict[node_id]
            node_element["fx"] = round(x_coord)
            node_element["fy"] = round(y_coord)
            node_element["fz"] = round(z_coord)

    def from_dict(self, data):
        """
        Method to build graph from dictionary.

        :param data: (dict) dictionary with nodes and link/edges details
        
        Sample data dictionary::

            sample_graph = {
                'nodes': [
                    {
                        'id': 'a',
                        'label': 'R1'
                    },
                    {
                        'id': 'b',
                        'label': 'somelabel',
                        'data': {'description': 'some node description'}
                    },
                    {
                        'id': 'e',
                        'label': 'E'
                    }
                ],
                'edges': [
                    {
                        'source': 'a',
                        'label': 'DF',
                        'src_label': 'Gi1/1',
                        'trgt_label': 'Gi2/2',
                        'target': 'b',
                        'url': 'google.com'
                    }
                ],
                'links': [
                    {
                        'source': 'a',
                        'target': 'e'
                    }
                ]
            }

        **Dictionary Content Rules**

        * dictionary may contain ``nodes`` key with a list of nodes dictionaries
        * each node dictionary must contain unique ``id`` attribute, other attributes are optional
        * dictionary may contain ``edges`` or ``links`` key with a list of edges dictionaries
        * each link dictionary must contain ``source`` and ``target`` attributes, other attributes are optional
        """
        for node in data.get("nodes", []):
            self.add_node(**node)
        for link in data.get("links", []):
            self.add_link(**link)
        for edge in data.get("edges", []):
            self.add_link(**edge)

    def from_list(self, data):
        """
        Method to build graph from list.

        :param data: (list) list of link dictionaries
        
        Sample list data::

            sample_graph = [
                {
                    'source': 'a',
                    'label': 'DF',
                    'src_label': 'Gi1/1',
                    'trgt_label': 'Gi2/2',
                    'target': 'b',
                    'data': {'vlans': 'vlans_trunked: 1,2,3\\nstate: up'}
                },
                {
                    'source': 'a',
                    'target': {
                            'id': 'e',
                            'label': 'somelabel',
                            'data': {'description': 'some node description'}
                        }
                    }
                }
            ]

        **List Content Rules**

        * each list item must have ``target`` and ``source`` attributes defined
        * ``target``/``source`` attributes can be either a string or a dictionary
        * dictionary ``target``/``source`` node must contain ``id`` attribute and
          other supported node attributes

        .. note:: By default drawio_diagram object ``node_duplicates`` action set to 'skip' 
            meaning that node will be added on first occurrence and ignored after that. Set 
            ``node_duplicates`` to 'update' if node with given id need to be updated by later 
            occurrences in the list.
        """
        for edge in data:
            self.add_link(**edge)

    def from_v3d_json(self, data):
        """
        Method to load `JSON input syntax <https://github.com/vasturiano/3d-force-graph#input-json-syntax>`_
        data into diagram plugin, presumably to perform various manipulations.
        
        :param data: (str) string of `JSON input syntax <https://github.com/vasturiano/3d-force-graph#input-json-syntax>`_ format
        """
        data_json = json.loads(data)

        # load nodes
        for node in data_json.get("nodes"):
            self.add_node(**node)

        # load links
        for link in data_json.get("links"):
            self.add_link(**link)

    def dump_dict(self):
        """
        Method to populate ``self.drawing`` dictionary with current links and nodes items,
        return ``self.drawing`` dictionary content after that.
        """
        self.drawing = {
            "nodes": list(self.nodes_dict.values()),
            "links": list(self.links_dict.values()),
        }
        return self.drawing

    def dump_json(self, **kwargs):
        """
        Method to transform graph data in a JSON formatted string.
        
        :param kwargs: (dict) kwargs to use with ``json.dumps`` method
        """
        gdict = self.dump_dict()
        return json.dumps(gdict, **kwargs)

    def dump_file(self, filename=None, folder="./Output/", json_kwargs=None):
        """
        Method to save current diagram to text file in a JSON format.

        :param filename: (str) name of the file to save diagram into
        :param folder: (str) OS path to folder where to save diagram file, default is ``./Output/``
        :param json_kwargs: (dict) kwargs to use with ``json.dumps`` method
        
        If no ``filename`` provided, timestamped format used to produce filename, 
        e.g.: ``Sun Jun 28 20-30-57 2020_output.txt``
        """
        import time

        json_kwargs = json_kwargs or {"sort_keys": True, "indent": 4}
        # create output folder if it does not exists
        os.makedirs(folder, exist_ok=True)
        # create file name
        if not filename:
            ctime = time.ctime().replace(":", "-")
            filename = "{}_output.txt".format(ctime)
        # save file to disk
        with open(os.path.join(folder, filename), "w") as outfile:
            outfile.write(self.dump_json(**json_kwargs))

    def run(
        self,
        ip: str = "0.0.0.0",
        port: int = 9000,
        debug: bool = True,
        dry_run: bool = False,
    ) -> None:
        """
        Method to run FLASK web server using built-in browser app.
        
        :param ip: IP address to bound WEB server to
        :param port: port number to run WEB server on
        :param debug: If True run Flask server in debug mode
        :param dry_run: (bool) if True, do not start, return status info instead
        """
        if dry_run:
            return {
                "message": "would start flask development server using graph_browser app",
                "ip": ip,
                "port": port,
            }

        from N2G.plugins.viewers.v3d_viewer import run_v3d_viewer

        run_v3d_viewer(ip=ip, port=port, debug=debug, diagram_data=self.dump_json())



"""
CLI IP Data Plugin
******************

This plugin populates diagram with IP related information, such as subnets and IP addresses.

IP data plugin mainly useful in networking domain, it can take show commands output from
network devices, parse it with TTP templates in a structure that processed further to
load into one of diagram plugin objects using ``from_dict`` method

Features Supported
------------------

**Support matrix**

+---------------+------------+-----------+-----------+-----------+-----------+
|  Platform     | IP/Subnets |   ARP     | interface | links     | FHRP      |
|  Name         |            |           | config    | grouping  | Protocols |
+===============+============+===========+===========+===========+===========+
| cisco_ios     |    YES     |    YES    |    YES    |    YES    |    YES    |
+---------------+------------+-----------+-----------+-----------+-----------+
| cisco_xr      |    YES     |    ---    |    YES    |    YES    |    YES    |
+---------------+------------+-----------+-----------+-----------+-----------+
| cisco_nxos    |    YES     |    YES    |    YES    |    YES    |    YES    |
+---------------+------------+-----------+-----------+-----------+-----------+
| huawei        |    YES     |    YES    |    YES    |    YES    |    YES    |
+---------------+------------+-----------+-----------+-----------+-----------+
| fortinet      |    YES     |    YES    |    YES    |    YES    |    ---    |
+---------------+------------+-----------+-----------+-----------+-----------+

Required Commands output
------------------------

cisco_ios:

* ``show running-configuration`` or ``show running-configuration | section interface`` - mandatory output, used to parse interfaces IP addresses
* ``show ip arp`` and/or ``show ip arp vrf xyz`` - required by ARP visualization feature

cisco_xr:

* ``show running-configuration`` or ``show running-configuration interface`` - mandatory output, used to parse interfaces IP addresses
* ``show arp`` and/or ``show arp vrf xyz/all`` - required by ARP visualization feature

cisco_nxos:

* ``show running-configuration`` or ``show running-configuration | section interface`` - mandatory output, used to parse interfaces IP addresses
* ``show ip arp`` - required by ARP visualization feature

huawei:

* ``display current-configuration interface`` - mandatory output, used to parse interfaces IP addresses
* ``display arp all`` - required by ARP visualization feature

fortinet:

* ``get system config``  - mandatory output, used to parse interfaces IP addresses
* ``get system arp`` - required by ARP visualization feature

Sample Usage
------------

Code to populate yEd diagram object with IP and subnet nodes using data dictionary::

    data = {"huawei": ['''
    <hua_sw1>dis current-configuration interface
    #
    interface Vlanif140
     ip binding vpn-instance VRF_MGMT
     ip address 10.1.1.2 255.255.255.0
     vrrp vrid 200 virtual-ip 10.1.1.1
    #
    interface Eth-Trunk5.123
     vlan-type dot1q 123
     description hua_sw2 BGP  peering
     ip binding vpn-instance VRF_MGMT
     ip address 10.0.0.1 255.255.255.252
     ipv6 address FD00:1::1/126
    #
    interface Eth-Trunk5.200
     vlan-type dot1q 200
     description hua_sw3 OSPF  peering
     ip address 192.168.2.2 255.255.255.252

    <hua_sw1>dis arp all
    10.1.1.2        a008-6fc1-1101        I         Vlanif140       VRF_MGMT
    10.1.1.1        a008-6fc1-1102   0    D         Vlanif140       VRF_MGMT
    10.1.1.3        a008-6fc1-1103   10   D/200     Vlanif140       VRF_MGMT
    10.1.1.9        a008-6fc1-1104   10   D/200     Vlanif140       VRF_MGMT
    10.0.0.2        a008-6fc1-1105   10   D/200     Eth-Trunk5.123  VRF_MGMT
        ''',
        '''
    <hua_sw2>dis current-configuration interface
    #
    interface Vlanif140
     ip binding vpn-instance VRF_MGMT
     ip address 10.1.1.3 255.255.255.0
     vrrp vrid 200 virtual-ip 10.1.1.1
    #
    interface Eth-Trunk5.123
     vlan-type dot1q 123
     description hua_sw1 BGP  peering
     ip binding vpn-instance VRF_MGMT
     ip address 10.0.0.2 255.255.255.252
     ipv6 address FD00:1::2/126
        ''',
        '''
    <hua_sw3>dis current-configuration interface
    #
    interface Vlanif200
     ip binding vpn-instance VRF_CUST1
     ip address 192.168.1.1 255.255.255.0
    #
    interface Eth-Trunk5.200
     vlan-type dot1q 200
     description hua_sw1 OSPF  peering
     ip address 192.168.2.1 255.255.255.252

    <hua_sw3>dis arp
    192.168.1.1         a008-6fc1-1111       I      Vlanif200
    192.168.1.10        a008-6fc1-1110   30  D/300  Vlanif200
        '''],
    "cisco_nxos": ['''
    switch_1# show run | sec interface
    interface Vlan133
      description OOB
      vrf member MGMT_OOB
      ip address 10.133.137.2/24
      hsrp 133
        preempt
        ip 10.133.137.1
    !
    interface Vlan134
      description OOB-2
      vrf member MGMT_OOB
      ip address 10.134.137.2/24
      vrrpv3 1334 address-family ipv4
        address 10.134.137.1 primary
    !
    interface Vlan222
      description PTP OSPF Routing pat to  siwtch2
      ip address 10.222.137.1/30
    !
    interface Vlan223
      description PTP OSPF Routing pat to siwtch3
      ip address 10.223.137.1/30

    switch_1# show ip arp vrf all
    10.133.137.2    -  d094.7890.1111  Vlan133
    10.133.137.1    -  d094.7890.1111  Vlan133
    10.133.137.30   -  d094.7890.1234  Vlan133
    10.133.137.91   -  d094.7890.4321  Vlan133
    10.134.137.1    -  d094.7890.1111  Vlan134
    10.134.137.2    -  d094.7890.1111  Vlan134
    10.134.137.3   90  d094.7890.2222  Vlan134
    10.134.137.31  91  d094.7890.beef  Vlan134
    10.134.137.81  81  d094.7890.feeb  Vlan134
    10.222.137.2   21  d094.7890.2222  Vlan222
    '''
    }

    drawing = create_yed_diagram()
    drawer = cli_ip_data(drawing, add_arp=True, add_fhrp=True)
    drawer.work(data)
    drawer.drawing.dump_file(filename="ip_graph_dc_1.graphml", folder="./Output/")

API Reference
-------------

.. autoclass:: N2G.plugins.data.cli_ip_data.cli_ip_data
   :members:
"""
import logging
import os
import json
import ipaddress

try:
    from ttp import ttp
    from ttp_templates import ttp_vars as ttp_templates_vars
    from ttp_templates import get_template

    HAS_TTP = True
except ImportError:
    HAS_TTP = False

# initiate logging
log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# Custom TTP functions to use in templates:
# -----------------------------------------------------------------------------


def add_network(data):
    if "netmask" in data:
        ip_obj, _ = _ttp_["match"]["to_ip"]("{}/{}".format(data["ip"], data["netmask"]))
    else:
        ip_obj, _ = _ttp_["match"]["to_ip"](data["ip"])
    data["network"] = str(ip_obj.network)
    data["netmask"] = str(ip_obj.network.prefixlen)
    return data, None


# -----------------------------------------------------------------------------
# Main class:
# -----------------------------------------------------------------------------


class cli_ip_data:
    """
    Class to instantiate IP Data Plugin.

    :param drawing: (obj) - N2G drawing object instantiated using drawing module e.g. yed_diagram or drawio_diagram
    :param ttp_vars: (dict) - dictionary to use as TTP parser object template variables
    :param platforms: (list) - list of platform names to process e.g. ``cisco_ios``, ``cisco_xr`` etc, default is ``_all_``
    :param group_links: (bool) - if True, will group links between same nodes, default is False
    :param add_arp: (bool) - if True, will add IP nodes from ARP parsing results, default is False
    :param label_interface: (bool) - if True, will add interface name to the link's source and target labels, default is False
    :param label_vrf: (bool) - if True, will add VRF name to the link's source and target labels, default is False
    :param collapse_ptp: (bool) - if True (default) combines links for ``/31`` and ``/30`` IPv4 and ``/127`` IPv6
      subnets into a single ink
    :param add_fhrp: (bool) - if True adds HSRP and VRRP IP addresses to the diagram, default is False
    :param bottom_label_length: (int) - bottom label length of interface description to use for subnet nodes,
      if False or 0, bottom label will not be set for subnet nodes
    :param lbl_next_to_subnet: (bool) - if True, put link ``port:vrf:ip`` label next to subnet node, default is False
    """

    def __init__(
        self,
        drawing,
        ttp_vars=None,
        group_links=False,
        add_arp=False,
        label_interface=False,
        label_vrf=False,
        collapse_ptp=True,
        add_fhrp=False,
        bottom_label_length=0,
        lbl_next_to_subnet=False,
        platforms=None,
    ):
        self.group_links = group_links
        self.add_arp = add_arp
        self.label_interface = label_interface
        self.label_vrf = label_vrf
        self.collapse_ptp = collapse_ptp
        self.add_fhrp = add_fhrp
        self.bottom_label_length = bottom_label_length
        self.lbl_next_to_subnet = lbl_next_to_subnet
        self.platforms = platforms or ["_all_"]
        self.ttp_vars = ttp_vars or {
            "IfsNormalize": ttp_templates_vars.short_interface_names
        }
        self.drawing = drawing
        self.drawing.node_duplicates = "update"
        self.parsed_data = {}
        self.nodes_dict = {}
        self.links_dict = {}
        self.graph_dict = {"nodes": [], "links": []}
        self.nodes_to_links_dict = {}  # used by group_links
        self.collapse_ptp_dict = {}  # used by collapse_ptp

    def _make_hash_tuple(self, item):
        """
        Method to create link hash tuple out of source, target, src_label
        and trgt_label values

        :param item: (dict) link dictionary
        """
        target = (
            item["target"]["id"] if isinstance(item["target"], dict) else item["target"]
        )
        return tuple(
            sorted(
                [
                    item["source"],
                    target,
                    item.get("src_label", ""),
                    item.get("trgt_label", ""),
                ]
            )
        )

    def work(self, data):
        """
        Method to parse text data and add nodes and links to drawing object.

        :param data: (dict or str) dictionary or OS path string to directories with data files

        If data is dictionary, keys must correspond to **Platform** column in
        `Features Supported`_ section table, values are lists of text items to
        process.

        Data dictionary sample::

            data = {
                "cisco_ios" : ["h1", "h2"],
                "cisco_xr": ["h3", "h4"],
                "cisco_nxos": ["h5", "h6"],
                ...etc...
            }

        Where ``hX`` devices show commands output.

        If data is an OS path directory string, child directories' names must correspond
        to **Platform** column in `Features Supported`_ section table. Each child directory
        should contain text files with show commands output for each device, names of files
        are arbitrary, but output should contain device prompt to extract device hostname.

        Directories structure sample::

            ├───folder_with_data
                ├───cisco_ios
                │       switch1.txt
                │       switch2.txt
                └───cisco_nxos
                        nxos_switch_1.txt
                        nxos_switch_2.txt

        To point N2G to above location ``data`` attribute string can be ``/var/data/n2g/folder_with_data/``
        """
        self._parse(data)
        self._form_base_graph_dict()
        # go through config statements
        if self.collapse_ptp:
            self._collapse_ptp()
        if self.group_links:
            self._group_links()
        # form graph dictionary and add it to drawing
        self._update_drawing()

    def _parse(self, data):
        """
        Function to parse text data using TTP templates

        :param data: (dict or str) data to parse
        """
        if not HAS_TTP:
            raise ModuleNotFoundError(
                "N2G:cli_ip_data failed importing TTP, is it installed?"
            )
        # process data dictionary
        if isinstance(data, dict):
            parser = ttp(vars=self.ttp_vars)
            parser.add_function(
                add_network, scope="group", name="add_network", add_ttp=True
            )
            for platform_name, text_list in data.items():
                if (
                    "_all_" not in self.platforms
                    and not platform_name in self.platforms
                ):
                    continue
                ttp_template = get_template(
                    misc="N2G/cli_ip_data/{}.txt".format(platform_name)
                )
                parser.add_template(template=ttp_template, template_name=platform_name)
                for item in text_list:
                    parser.add_input(item, template_name=platform_name)
        # process directories at OS path
        elif isinstance(data, str):
            parser = ttp(vars=self.ttp_vars, base_path=data)
            parser.add_function(
                add_network, scope="group", name="add_network", add_ttp=True
            )
            # get all sub-folders and load respective templates
            with os.scandir(data) as dirs:
                for entry in dirs:
                    if entry.is_dir():
                        platform_name = entry.name
                        if (
                            "_all_" not in self.platforms
                            and not platform_name in self.platforms
                        ):
                            continue
                        ttp_template = get_template(
                            misc="N2G/cli_ip_data/{}.txt".format(platform_name)
                        )
                        parser.add_template(
                            template=ttp_template, template_name=entry.name
                        )
                        parser.add_input(
                            data=os.path.abspath(entry), template_name=platform_name
                        )
        else:
            log.error(
                "Expecting dictionary or string, but '{}' given".format(type(data))
            )
            return
        parser.parse(one=True)
        self.parsed_data = parser.result(structure="dictionary")
        # pprint.pprint(self.parsed_data, width = 100)

    def _form_base_graph_dict(self):
        """
        Method to form graph dictionaries of nodes and links out of parsing results.
        """
        interfaces_ip = {}  # need this dict to skip ARP entries
        for platform, hosts_data in self.parsed_data.items():
            for hostname, host_data in hosts_data.items():
                self._add_node({"id": hostname, "top_label": "Device"}, host_data)
                for interface, interface_data in host_data["interfaces"].items():
                    interface_networks = []
                    for ip in interface_data.get("ip_addresses", []):
                        network = ip["network"]
                        if self.add_arp or self.add_fhrp:
                            interface_networks.append(network)
                            interfaces_ip.setdefault(network, []).append(ip["ip"])
                        network_node = {"id": network, "top_label": "Subnet"}
                        # add bottom lable to node
                        if (
                            interface_data.get("port_description")
                            and self.bottom_label_length
                        ):
                            if (
                                len(interface_data["port_description"])
                                > self.bottom_label_length
                            ):
                                network_node["bottom_label"] = "{}..".format(
                                    interface_data["port_description"][
                                        : self.bottom_label_length
                                    ]
                                )
                            else:
                                network_node["bottom_label"] = "{}".format(
                                    interface_data["port_description"]
                                )
                        link_description_data = {
                            k: v
                            for k, v in interface_data.items()
                            if not k in ["arp", "ip_addresses"]
                        }
                        link_description_data.update(ip)
                        link_description = {
                            "{}:{}".format(hostname, interface): link_description_data
                        }
                        link_dict = {
                            "source": network,
                            "target": hostname,
                            "description": json.dumps(
                                link_description,
                                sort_keys=True,
                                indent=4,
                                separators=(",", ": "),
                            ),
                        }
                        if ip["netmask"] in ["32", "128"]:
                            network_node["top_label"] = interface
                            network_node["bottom_label"] = interface_data.get(
                                "vrf", "global"
                            )
                        else:
                            link_label = "{}/{}".format(ip["ip"], ip["netmask"])
                            if self.label_vrf:
                                link_label = "{}:{}".format(
                                    interface_data.get("vrf", "global"), link_label
                                )
                            if self.label_interface:
                                link_label = "{}:{}".format(interface, link_label)
                            if self.lbl_next_to_subnet:
                                link_dict["src_label"] = link_label
                            else:
                                link_dict["trgt_label"] = link_label
                        # add new node and link to graph
                        self._add_node(network_node)
                        self._add_link(link_dict, network)
                    # check if need to add FHRP IPs
                    if self.add_fhrp:
                        interface_network_objects = [
                            ipaddress.ip_network(i) for i in interface_networks
                        ]
                        for fhrp_entry in interface_data.get("fhrp", []):
                            # get ip entry network
                            ip = fhrp_entry["ip"]
                            ip_obj = ipaddress.ip_address(ip)
                            network = str(
                                [i for i in interface_network_objects if ip_obj in i][0]
                            )
                            # add IP address node
                            node_id = "{}:{}".format(network, ip)
                            description = {
                                "FHRP:{}:{}".format(hostname, interface): fhrp_entry
                            }
                            self._add_node(
                                {
                                    "id": node_id,
                                    "top_label": "{} VIP".format(fhrp_entry["type"]),
                                    "label": ip,
                                    "description": json.dumps(
                                        description,
                                        sort_keys=True,
                                        indent=4,
                                        separators=(",", ": "),
                                    ),
                                }
                            )
                            # add link to network
                            link_dict = {"source": node_id, "target": network}
                            self._add_link(link_dict)
                    # check if need to add ARP to diagram
                    if self.add_arp:
                        interface_network_objects = [
                            ipaddress.ip_network(i) for i in interface_networks
                        ]
                        for arp_entry in interface_data.get("arp", []):
                            # get arp entry network
                            ip = arp_entry["ip"]
                            ip_obj = ipaddress.ip_address(ip)
                            network = str(
                                [i for i in interface_network_objects if ip_obj in i][0]
                            )
                            # add IP address node
                            node_id = "{}:{}".format(network, ip)
                            description = {
                                "ARP:{}:{}".format(hostname, interface): arp_entry
                            }
                            self._add_node(
                                {
                                    "id": node_id,
                                    "top_label": "ARP entry",
                                    "label": ip,
                                    "description": json.dumps(
                                        description,
                                        sort_keys=True,
                                        indent=4,
                                        separators=(",", ": "),
                                    ),
                                }
                            )
                            # add link to network
                            link_dict = {"source": node_id, "target": network}
                            self._add_link(link_dict)
        # clean up ARP entries that duplicate interface IPs
        if self.add_arp:
            for network, ips in interfaces_ip.items():
                for ip in ips:
                    arp_node_id = "{}:{}".format(network, ip)
                    link_hash = self._make_hash_tuple(
                        {"source": arp_node_id, "target": network}
                    )
                    _ = self.nodes_dict.pop(arp_node_id, None)
                    _ = self.links_dict.pop(link_hash, None)

    def _add_node(self, item, host_data=None):
        """
        Method to add single node to nodes dictionary if it does not exist or
        update existing node.

        :param item: (dict) node dictionary to process
        :param host_data: (dict) dictionary with network device details
        """
        host_data = host_data or {}
        # add new node
        if not item["id"] in self.nodes_dict:
            if host_data.get("node_facts"):
                item["description"] = json.dumps(
                    host_data["node_facts"],
                    sort_keys=True,
                    indent=4,
                    separators=(",", ": "),
                )
            self.nodes_dict[item["id"]] = item
        # update node attributes if they do not exists already
        else:
            node = self.nodes_dict[item["id"]]
            for key, value in item.items():
                if not key in node:
                    node[key] = value
            if not "description" in node and host_data.get("node_facts"):
                node["description"] = json.dumps(
                    host_data["node_facts"],
                    sort_keys=True,
                    indent=4,
                    separators=(",", ": "),
                )
            # merge descriptions
            if "description" in node and "description" in item:
                node["description"] = json.loads(node["description"])
                item["description"] = json.loads(item["description"])
                node["description"].update(item["description"])
                node["description"] = json.dumps(
                    node["description"],
                    sort_keys=True,
                    indent=4,
                    separators=(",", ": "),
                )

    def _add_link(self, item, network=None):
        """
        Method to add single link to links dictionary.

        :param item: (dict) link dictionary to process
        :param network: (str) link subnet value e.g. "10.0.0.0/30"
        """
        link_hash = self._make_hash_tuple(item)
        if link_hash not in self.links_dict:
            self.links_dict[link_hash] = item
        # check if need to pre-process nodes_to_links_dict used by group_links
        if self.group_links:
            self._update_nodes_to_links_dict(item, link_hash)
        if self.collapse_ptp and network:
            if (network.split("/")[1] in ["30", "31"] and "." in network) or (
                network.split("/")[1] in ["127"] and ":" in network
            ):
                self.collapse_ptp_dict.setdefault(network, []).append(link_hash)

    def _update_nodes_to_links_dict(self, item, link_hash):
        """
        Method to update nodes_to_links_dict, used by group_links feature.

        :param item: (dict) link dictionary
        :param link_hash: (tuple) link identifier hash tuple
        """
        src = item["source"]
        tgt = item["target"]
        nodes_hash = tuple(sorted([src, tgt]))
        self.nodes_to_links_dict.setdefault(nodes_hash, [])
        if not link_hash in self.nodes_to_links_dict[nodes_hash]:
            self.nodes_to_links_dict[nodes_hash].append(link_hash)

    def _group_links(self):
        """
        Method to group links between nodes and update links_dict
        """
        # find nodes that have more then 1 link in between, group links
        for node_pair, link_hashes in self.nodes_to_links_dict.items():
            if len(link_hashes) < 2:
                continue
            grouped_link = {"source": node_pair[0], "target": node_pair[1]}
            description = {"grouped_links": {}}
            links_to_group_count = 0
            for link_hash in link_hashes:
                if link_hash in self.links_dict:
                    links_to_group_count += 1
                    link_data = self.links_dict[link_hash]
                    src_label = "{}:{}".format(
                        link_data["source"], link_data.get("src_label", "")
                    )
                    trgt_label = "{}:{}".format(
                        link_data["target"], link_data.get("trgt_label", "")
                    )
                    description["grouped_links"][src_label] = trgt_label
                    description["link-{}".format(links_to_group_count)] = json.loads(
                        link_data["description"]
                    )
            # skip grouping links that does not have any members left in self.links_dict
            if links_to_group_count >= 2:
                # remove grouped links from links_dict
                for link_hash in link_hashes:
                    _ = self.links_dict.pop(link_hash, None)
                # form grouped links description and label
                grouped_link["description"] = json.dumps(
                    description, sort_keys=True, indent=4, separators=(",", ": ")
                )
                grouped_link["label"] = "x{}".format(links_to_group_count)
                grouped_link_hash = self._make_hash_tuple(grouped_link)
                self.links_dict[grouped_link_hash] = grouped_link
        del self.nodes_to_links_dict

    def _collapse_ptp(self):
        """
        Method to combine ptp link into a single link, by default this plugin adds ptp
        subnets as nodes with links to devices that have IP addresses out of that subnet,
        this method deleted ptp subnet node and adds a link between devices instead.

        self.collapse_ptp_dict - mapping of ptp networks to link hashes used to combine
        ptp links.

        Sample::

            {'10.123.2.2/31': [('', '10.123.2.2/31', '10.123.2.3/31', 'switch_1'),
                            ('', '10.123.2.2/31', '10.123.2.2/31', 'switch_2')],
            '10.123.2.4/31': [('', '10.123.2.4/31', '10.123.2.4/31', 'switch_1'),
                            ('', '10.123.2.4/31', '10.123.2.5/31', 'switch_2')]}
        """
        for network, links in self.collapse_ptp_dict.items():
            if not len(links) == 2:
                continue
            # remove ptp network node
            _ = self.nodes_dict.pop(network)
            # remove old links to ptp network
            # and create new link between devices
            link_1 = self.links_dict.pop(links[0])
            link_2 = self.links_dict.pop(links[1])
            description = {
                **json.loads(link_1.get("description", {})),
                **json.loads(link_2.get("description", {})),
            }
            link_dict = {
                "source": link_1["target"],
                "target": link_2["target"],
                "src_label": link_1["trgt_label"]
                if link_1.get("trgt_label")
                else link_1["src_label"],
                "trgt_label": link_2["trgt_label"]
                if link_2.get("trgt_label")
                else link_2["src_label"],
                "description": json.dumps(
                    description, sort_keys=True, indent=4, separators=(",", ": ")
                ),
            }
            self._add_link(link_dict)

    def _update_drawing(self):
        """
        Method to populate drawing object with processed nodes and links using from_dict method.
        """
        self.graph_dict["nodes"] = list(self.nodes_dict.values())
        self.graph_dict["links"] = list(self.links_dict.values())
        # pprint.pprint(self.graph_dict, width =100)
        self.drawing.from_dict(self.graph_dict)



import xml.etree.ElementTree as ET
import os
import hashlib
import logging

# initiate logging
log = logging.getLogger(__name__)


class drawio_diagram:
    """
    N2G DrawIO module allows to produce diagrams compatible with DrawIO XML format.

    **Parameters**

    * ``node_duplicates`` (str) can be of value skip, log, update
    * ``link_duplicates`` (str) can be of value skip, log, update

    """

    # XML string templates to create etree elements from:
    drawio_drawing_xml = """
    <mxfile type="device" compressed="false">
    </mxfile>
    """

    drawio_diagram_xml = """
    <diagram id="{id}" name="{name}">
      <mxGraphModel dx="{width}" dy="{height}" grid="1" gridSize="10" guides="1" tooltips="1" connect="1" arrows="1" fold="1" page="1" pageScale="1" pageWidth="827" pageHeight="1169" math="0" shadow="1">
        <root>
          <mxCell id="0"/>   
          <mxCell id="1" parent="0"/>
        </root>
      </mxGraphModel>
    </diagram>
    """

    drawio_node_object_xml = """
    <object id="{id}" label="{label}">
      <mxCell style="{style}" vertex="1" parent="1">
          <mxGeometry x="{x_pos}" y="{y_pos}" width="{width}" height="{height}" as="geometry"/>
      </mxCell>
    </object>
    """

    drawio_link_object_xml = """
    <object id="{id}" label="{label}">
      <mxCell style="{style}" edge="1" parent="1" source="{source_id}" target="{target_id}">
          <mxGeometry relative="1" as="geometry"/>
      </mxCell>
    </object>
    """

    drawio_link_label_xml = """
    <mxCell id="{id}" value="{label}" style="{style};" vertex="1" connectable="0" parent="{parent_id}">
      <mxGeometry x="{x}" relative="{rel}" as="geometry">
        <mxPoint as="offset" />
      </mxGeometry>
    </mxCell>
    """

    def __init__(self, node_duplicates="skip", link_duplicates="skip"):
        self.drawing = ET.fromstring(self.drawio_drawing_xml)
        self.nodes_ids = {}  # dictionary of {diagram_name: [node_id1, node_id2]}
        self.edges_ids = {}  # dictionary of {diagram_name: [edge_id1, edge_id2]}
        self.node_duplicates = node_duplicates
        self.link_duplicates = link_duplicates
        self.current_diagram = None
        self.current_diagram_id = ""
        self.default_node_style = "rounded=1;whiteSpace=wrap;html=1;"
        self.default_link_style = "endArrow=none;"
        self.default_link_label_style = "labelBackgroundColor=#ffffff;"

    def add_diagram(self, id, name="", width=1360, height=864):
        """
        Method to add new diagram tab and switch to it.

        .. warning:: This method must be called to create at list one
          diagram tab to work with prior to nodes and links can be added
          to the drawing calling ``add_link`` or ``add_node`` methods.

        **Parameters**

        * ``id`` (str) id of the diagram, should be unique across other diagrams
        * ``name`` (str) tab name
        * ``width`` (int) width of diagram in pixels
        * ``height`` (int) height of diagram in pixels

        """
        if id in self.nodes_ids or id in self.edges_ids:
            return
        if not name.strip():
            name = id
        diagram = ET.fromstring(
            self.drawio_diagram_xml.format(id=id, name=name, width=width, height=height)
        )
        self.nodes_ids[id] = []
        self.edges_ids[id] = []
        self.drawing.append(diagram)
        self.go_to_diagram(diagram_name=name)

    def go_to_diagram(self, diagram_name=None, diagram_index=None):
        """
        DrawIO supports adding multiple diagram tabs within single document.
        This method allows to switch between diagrams in different tabs. That
        way each tab can be updated separately.

        **Parameters**

        * ``diagram_name`` (str) name of diagram tab to switch to
        * ``diagram_index`` (int) index of diagram tab to switch to, will
          change to last tab if index is out of range. Index can be positive
          or negative number and follows Python list index behavior. For
          instance, index equal to "-1" we go to last tab, "0" will go to
          first tab

        """
        if diagram_name is not None:
            self.current_diagram = self.drawing.find(
                "./diagram[@name='{name}']".format(name=diagram_name)
            )
        elif diagram_index is not None:
            try:
                self.current_diagram = self.drawing.findall("./diagram")[diagram_index]
            except IndexError:
                self.current_diagram = self.drawing.findall("./diagram")[-1]
        self.current_root = self.current_diagram.find("./mxGraphModel/root")
        self.current_diagram_id = self.current_diagram.attrib["id"]

    def _add_data_or_url(self, element, data, url):
        # add data if any
        attribs = {k: str(v) for k, v in data.items()}
        # add URL if any
        if url:
            # check if url is another diagram name
            diagram_link = self.drawing.find("./diagram[@name='{}']".format(url))
            if diagram_link is not None:
                url = "data:page/id,{diagram_id}".format(
                    diagram_id=diagram_link.attrib["id"]
                )
            attribs["link"] = url
        element.attrib.update(attribs)
        return element

    def _node_exists(self, id, **kwargs):
        # check if node with given id already exists
        if id in self.nodes_ids[self.current_diagram_id]:
            if self.node_duplicates == "log":
                log.error("add_shape_node: node '{}' already added to graph".format(id))
            elif self.node_duplicates == "skip":
                pass
            elif self.node_duplicates == "update":
                self.update_node(id, **kwargs)
            return True
        return False

    def add_node(
        self,
        id,
        label="",
        data=None,
        url="",
        style="",
        width=120,
        height=60,
        x_pos=200,
        y_pos=150,
        **kwargs
    ):
        """
        Method to add node to the diagram.

        **Parameters**

        * ``id`` (str) mandatory, unique node identifier, usually equal to node name
        * ``label`` (str) node label, if not provided, set equal to id
        * ``data`` (dict) dictionary of key value pairs to add as node data
        * ``url`` (str) url string to save as node ``url`` attribute
        * ``width`` (int) node width in pixels
        * ``height`` (int) node height in pixels
        * ``x_pos`` (int) node position on x axis
        * ``y_pos`` (int) node position on y axis
        * ``style`` (str) string containing DrawIO style parameters to apply to the node

        Sample DrawIO style string for the node::

            shape=mxgraph.cisco.misc.asr_1000_series;html=1;pointerEvents=1;
            dashed=0;fillColor=#036897;strokeColor=#ffffff;strokeWidth=2;
            verticalLabelPosition=bottom;verticalAlign=top;align=center;
            outlineConnect=0;

        """
        data = data or {}
        node_data = {}
        if self._node_exists(id, label=label, data=data, url=url):
            return
        self.nodes_ids[self.current_diagram_id].append(id)
        if not label.strip():
            label = id
        # try to get style from file
        if os.path.isfile(style[:5000]):
            with open(style, "r") as style_file:
                style = style_file.read()
        # create node element
        node = ET.fromstring(
            self.drawio_node_object_xml.format(
                id=id,
                label=label,
                width=width if str(width).strip() else 120,
                height=height if str(height).strip() else 60,
                x_pos=x_pos,
                y_pos=y_pos,
                style=style if style else self.default_node_style,
            )
        )
        # add data attributes and/or url to node
        node_data.update(data)
        node_data.update(kwargs)
        node = self._add_data_or_url(node, node_data, url)
        self.current_root.append(node)

    def update_node(
        self,
        id,
        label=None,
        data=None,
        url=None,
        style="",
        width="",
        height="",
        **kwargs
    ):
        """
        Method to update node details. Uses node ``id`` to search for node to update

        **Parameters**

        * ``id`` (str) mandatory, unique node identifier
        * ``label`` (str) label at the center of the node
        * ``data`` (dict) dictionary of data items to add to the node
        * ``width`` (int) node width in pixels
        * ``height`` (int) node height in pixels
        * ``url`` (str) url string to save as node `url` attribute
        * ``style`` (str) string containing DrawIO style parameters to apply to the node

        """
        data = data or {}
        node_data = {}
        node = self.current_root.find("./object[@id='{}']".format(id))
        # update data and url attributes
        node_data.update(data)
        node_data.update(kwargs)
        node = self._add_data_or_url(node, node_data, url)
        # update label
        if not label is None:
            node.attrib["label"] = label
        # update style
        mxCell_elem = node.find("./mxCell")
        if os.path.isfile(style[:5000]):
            with open(style, "r") as style_file:
                mxCell_elem.attrib["style"] = style_file.read()
        elif style:
            mxCell_elem.attrib["style"] = style
        # update size
        mxGeometry_elem = node.find("./mxCell/mxGeometry")
        if width:
            mxGeometry_elem.attrib["width"] = str(width)
        if height:
            mxGeometry_elem.attrib["height"] = str(height)

    def _link_exists(self, id, edge_tup):
        """method, used to check duplicate edges"""
        # check if edge with given id already exists
        if id in self.edges_ids[self.current_diagram_id]:
            if self.link_duplicates == "log":
                log.error(
                    "_link_exists: edge '{}' already added to graph".format(
                        ",".join(edge_tup)
                    )
                )
            elif self.link_duplicates == "skip":
                pass
            return True
        self.edges_ids[self.current_diagram_id].append(id)

    def add_link(
        self,
        source,
        target,
        style="",
        label="",
        data=None,
        url="",
        src_label="",
        trgt_label="",
        src_label_style="",
        trgt_label_style="",
        link_id=None,
        **kwargs
    ):
        """
        Method to add link between nodes to the diagram.

        **Parameters**

        * ``source`` (str) mandatory, source node id
        * ``source`` (str) mandatory, target node id
        * ``label`` (str) link label to display at the center of the link
        * ``data`` (dict) dictionary of key value pairs to add as link data
        * ``url`` (str) url string to save as link ``url`` attribute
        * ``style`` (str) string or OS path to text file with style to apply to the link
        * ``src_label`` (str) link label to display next to source node
        * ``trgt_label`` (str) link label to display next to target node
        * ``src_label_style`` (str) source label style string
        * ``trgt_label_style`` (str) target label style string
        * ``link_id`` (str or int) optional link id value, must be unique across all links

        Sample DrawIO style string for the link::

            endArrow=classic;fillColor=#f8cecc;strokeColor=#FF3399;dashed=1;
            edgeStyle=entityRelationEdgeStyle;startArrow=diamondThin;startFill=1;
            endFill=0;strokeWidth=5;

        .. note:: If source or target nodes does not exists, they will be automatically
          created

        All labels are optional and substituted with empty values to calculate link id.
        """
        data = data or {}
        link_data = {}
        # check type of source and target attribute
        source_node_dict = source.copy() if isinstance(source, dict) else {"id": source}
        source = source_node_dict.pop("id")
        target_node_dict = target.copy() if isinstance(target, dict) else {"id": target}
        target = target_node_dict.pop("id")
        # check if target and source nodes exist, add it if not,
        # self._node_exists method will update node
        # if self.node_duplicates set to update, by default its set to skip
        if not self._node_exists(source, **source_node_dict):
            self.add_node(id=source, **source_node_dict)
        if not self._node_exists(target, **target_node_dict):
            self.add_node(id=target, **target_node_dict)
        # create link id if not given
        if link_id:
            link_id = "link_id:{}".format(link_id)
            edge_tup = link_id
        else:
            edge_tup = tuple(sorted([label, source, target, src_label, trgt_label]))
            link_id = hashlib.md5(",".join(edge_tup).encode()).hexdigest()
        if self._link_exists(link_id, edge_tup):
            return
        # try to get style from file
        if os.path.isfile(style[:5000]):
            with open(style, "r") as style_file:
                style = style_file.read()
        # create link
        link = ET.fromstring(
            self.drawio_link_object_xml.format(
                id=link_id,
                label=label,
                source_id=source,
                target_id=target,
                style=style or self.default_link_style,
            )
        )
        # add link source label
        if src_label:
            src_label_obj = ET.fromstring(
                self.drawio_link_label_xml.format(
                    id="{}-src".format(link_id),
                    label=src_label,
                    parent_id=link_id,
                    style=src_label_style or self.default_link_label_style,
                    x="-0.5",
                    rel="1",
                )
            )
            self.current_root.append(src_label_obj)
            kwargs["src_label"] = src_label
        # add link target label
        if trgt_label:
            trgt_label_obj = ET.fromstring(
                self.drawio_link_label_xml.format(
                    id="{}-trgt".format(link_id),
                    label=trgt_label,
                    parent_id=link_id,
                    style=trgt_label_style or self.default_link_label_style,
                    x="0.5",
                    rel="-1",
                )
            )
            self.current_root.append(trgt_label_obj)
            kwargs["trgt_label"] = trgt_label
        # add links data and url
        link_data.update(data)
        link_data.update(kwargs)
        link_data.update({"source": source, "target": target})
        link = self._add_data_or_url(link, link_data, url)
        # save link to graph
        self.current_root.append(link)

    def dump_xml(self):
        """
        Method to return current diagram XML text
        """
        ret = ET.tostring(self.drawing, encoding="unicode")
        return ret

    def dump_file(self, filename=None, folder="./Output/"):
        """
        Method to save current diagram in .drawio file.

        **Parameters**

        * ``filename`` (str) name of the file to save diagram into
        * ``folder`` (str) OS path to folder where to save diagram file

        If no ``filename`` provided, timestamped format will be
        used to produce filename, e.g.: ``Sun Jun 28 20-30-57 2020_output.drawio``

        """
        import time

        # create output folder if it does not exists
        os.makedirs(folder, exist_ok=True)
        # create file name
        if not filename:
            ctime = time.ctime().replace(":", "-")
            filename = "{}_output.drawio".format(ctime)
        # save file to disk
        with open(os.path.join(folder, filename), "w") as outfile:
            outfile.write(self.dump_xml())

    def layout(self, algo="kk", **kwargs):
        """
        Method to calculate graph layout using Python
        `igraph <https://igraph.org/python/doc/tutorial/tutorial.html#layout-algorithms>`_
        library

        **Parameters**

        * ``algo`` (str) name of layout algorithm to use, default is 'kk'. Reference
          `Layout algorithms` table below for valid algo names
        * ``kwargs`` any additional kwargs to pass to igraph ``Graph.layout`` method

        **Layout algorithms**

        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | algo name                       |    description                                                                                                 |
        +=================================+================================================================================================================+
        | circle, circular                | Deterministic layout that places the vertices on a circle                                                      |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | drl                             | The Distributed Recursive Layout algorithm for large graphs                                                    |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | fr                              | Fruchterman-Reingold force-directed algorithm                                                                  |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | fr3d, fr_3d                     | Fruchterman-Reingold force-directed algorithm in three dimensions                                              |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | grid_fr                         | Fruchterman-Reingold force-directed algorithm with grid heuristics for large graphs                            |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | kk                              | Kamada-Kawai force-directed algorithm                                                                          |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | kk3d, kk_3d                     | Kamada-Kawai force-directed algorithm in three dimensions                                                      |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | large, lgl, large_graph         | The Large Graph Layout algorithm for large graphs                                                              |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | random                          | Places the vertices completely randomly                                                                        |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | random_3d                       | Places the vertices completely randomly in 3D                                                                  |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | rt, tree                        | Reingold-Tilford tree layout, useful for (almost) tree-like graphs                                             |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | rt_circular, tree               | Reingold-Tilford tree layout with a polar coordinate post-transformation, useful for (almost) tree-like graphs |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        | sphere, spherical, circular_3d  | Deterministic layout that places the vertices evenly on the surface of a sphere                                |
        +---------------------------------+----------------------------------------------------------------------------------------------------------------+
        """
        try:
            from igraph import Graph as ig
        except ImportError:
            raise SystemExit(
                "Failed to import igraph, install - pip install python-igraph"
            )
        # iterate over diagrams and layout elements
        for diagram in self.drawing.findall("./diagram"):
            igraph_graph = ig()
            self.go_to_diagram(diagram.attrib["name"])
            # populate igraph with nodes and edges from object tags
            for item in self.current_root.iterfind("./object"):
                # add edges, item[0] refernece to object's mxCell child tag
                if item[0].get("source") and item[0].get("target"):
                    igraph_graph.add_vertex(name=item[0].get("source"))
                    igraph_graph.add_vertex(name=item[0].get("target"))
                    igraph_graph.add_edge(
                        source=item[0].get("source"), target=item[0].get("target")
                    )
                # add nodes
                else:
                    igraph_graph.add_vertex(name=item.get("id"))
            # calculate layout
            layout = igraph_graph.layout(layout=algo, **kwargs)
            # scale layout to diagram size
            w = int(self.current_diagram.find("./mxGraphModel").attrib["dx"])
            h = int(self.current_diagram.find("./mxGraphModel").attrib["dy"])
            layout.fit_into(bbox=(w, h))
            # add coordinates from layout to diagram nodes
            for index, coord_item in enumerate(layout.coords):
                x_coord, y_coord = coord_item
                node_id = igraph_graph.vs[index].attributes()["name"]
                node_geometry_element = self.current_root.find(
                    "./object[@id='{id}']/mxCell/mxGeometry".format(id=node_id)
                )
                node_geometry_element.set("x", str(round(x_coord)))
                node_geometry_element.set("y", str(round(y_coord)))

    def from_dict(self, data, diagram_name="Page-1", width=1360, height=864):
        """
        Method to build graph from dictionary.

        **Parameters**

        * ``diagram_name`` (str) name of the diagram tab where to add links and nodes.
          Diagram tab will be created if it does not exists
        * ``width`` (int) diagram width in pixels
        * ``height`` (int) diagram height in pixels
        * ``data`` (dict) dictionary with nodes and link/edges details, example::

            sample_graph = {
                'nodes': [
                    {
                        'id': 'a',
                        'label': 'R1'
                    },
                    {
                        'id': 'b',
                        'label': 'somelabel',
                        'data': {'description': 'some node description'}
                    },
                    {
                        'id': 'e',
                        'label': 'E'
                    }
                ],
                'edges': [
                    {
                        'source': 'a',
                        'label': 'DF',
                        'src_label': 'Gi1/1',
                        'trgt_label': 'Gi2/2',
                        'target': 'b',
                        'url': 'google.com'
                    }
                ],
                'links': [
                    {
                        'source': 'a',
                        'target': 'e'
                    }
                ]
            }

        **Dictionary Content Rules**

        * dictionary may contain ``nodes`` key with a list of nodes dictionaries
        * each node dictionary must contain unique ``id`` attribute, other attributes are optional
        * dictionary may contain ``edges`` or ``links`` key with a list of edges dictionaries
        * each link dictionary must contain ``source`` and ``target`` attributes, other attributes are optional

        """
        self.add_diagram(id=diagram_name, width=width, height=height)
        for node in data.get("nodes", []):
            self.add_node(**node)
        # add links
        for link in data.get("links", []):
            link["source"] = link["source"].replace("\n", "&lt;br&gt;")
            link["target"] = link["target"].replace("\n", "&lt;br&gt;")
            link["label"] = link.get("label", "").replace("\n", "&#xa;")
            link["src_label"] = link.get("src_label", "").replace("\n", "&#xa;")
            link["trgt_label"] = link.get("trgt_label", "").replace("\n", "&#xa;")
            self.add_link(**link)
        # add links
        for edge in data.get("edges", []):
            self.add_link(**edge)

    def from_list(self, data, diagram_name="Page-1", width=1360, height=864):
        """
        Method to build graph from list.

        **Parameters**

        * ``diagram_name`` (str) name of the diagram tab where to add links and nodes.
          Diagram tab will be created if it does not exists
        * ``width`` (int) diagram width in pixels
        * ``height`` (int) diagram height in pixels
        * ``data`` (list) list of link dictionaries, example::

            sample_graph = [
                {
                    'source': 'a',
                    'label': 'DF',
                    'src_label': 'Gi1/1',
                    'trgt_label': 'Gi2/2',
                    'target': 'b',
                    'data': {'vlans': 'vlans_trunked: 1,2,3\\nstate: up'}
                },
                {
                    'source': 'a',
                    'target': {
                            'id': 'e',
                            'label': 'somelabel',
                            'data': {'description': 'some node description'}
                        }
                    }
                }
            ]

        **List Content Rules**

            * each list item must have ``target`` and ``source`` attributes defined
            * ``target``/``source`` attributes can be either a string or a dictionary
            * dictionary ``target``/``source`` node must contain ``id`` attribute and
              other supported node attributes

        .. note::

            By default drawio_diagram object ``node_duplicates`` action set to 'skip' meaning that node will be added on first occurrence
            and ignored after that. Set ``node_duplicates`` to 'update' if node with given id need to be updated by
            later occurrences in the list.
        """
        self.add_diagram(id=diagram_name, width=width, height=height)
        [self.add_link(**edge) for edge in data]

    def from_file(self, filename, file_load="xml"):
        """
        Method to load nodes and links from Drawio diagram file for
        further processing

        **Args**

            * filename - OS path to .drawio file to load
        """
        with open(filename, "r") as f:
            if file_load == "xml":
                self.from_xml(f.read())

    def from_xml(self, text_data):
        """
        Method to load graph from .drawio XML text produced by DrawIO

        **Args**

            * text_data - text data to load
        """
        self.drawing = ET.fromstring(text_data)
        # extract diagrams, nodes, edges IDs
        for diagram_elem in self.drawing.findall("./diagram"):
            self.nodes_ids.setdefault(diagram_elem.attrib["id"], [])
            self.edges_ids.setdefault(diagram_elem.attrib["id"], [])
            # iterate over mxcells to extract nodes and edges
            for object in diagram_elem.findall("./mxGraphModel/root/object"):
                object_id = object.attrib["id"]
                mxCell = object.find("./mxCell")
                # check if this is the edge
                if "source" in mxCell.attrib and "target" in mxCell.attrib:
                    self.edges_ids[diagram_elem.attrib["id"]].append(object_id)
                else:
                    self.nodes_ids[diagram_elem.attrib["id"]].append(object_id)
        self.go_to_diagram(diagram_index=0)

    def from_csv(self, text_data):
        """
        Method to build graph from CSV tables

        **Parameters**

        * ``text_data`` (str) CSV text with links or nodes details

        This method supports loading CSV text data that contains nodes or links
        information. If ``id`` in headers, ``from_dict`` method will be called for CSV
        processing, ``from_list`` method will be used otherwise.

        CSV data with nodes details should have headers matching ``add_node`` method
        arguments and rules.

        CSV data with links details should have headers matching ``add_link`` method
        arguments and rules.

        Sample CSV table with links details::

            "source","label","target","src_label","trgt_label"
            "a","DF","b","Gi1/1","Gi2/2"
            "b","Copper","c","Te2/1",
            "b","Copper","e","","GE3"
            "d","FW","e",,

        Sample CSV table with nodes details::

            "id","label","style","width","height"
            a,"R1,2","./Pics/cisco_router.txt",78,53
            "b","some",,,
            "c","somelabel",,,
            "d","somelabel1",,,
            "e","R1",,,

        """
        # import libs
        from io import StringIO
        import csv

        # need to handle text data as file like object for csv reader to work
        iostring = StringIO(newline="")
        iostring.write(text_data)
        iostring.seek(0)
        # load csv data
        dict_reader = csv.DictReader(iostring)
        data_list = list(dict_reader)
        # if id given - meaning it is nodes data
        if data_list[0].get("id"):
            self.from_dict({"nodes": data_list})
        else:
            self.from_list(data_list)

    def update_link(
        self,
        edge_id="",
        label="",
        source="",
        target="",
        data=None,
        url="",
        style="",
        src_label="",
        trgt_label="",
        new_label=None,
        new_src_label=None,
        new_trgt_label=None,
        src_label_style="",
        trgt_label_style="",
        **kwargs
    ):
        """
        Method to update edge/link details.

        **Parameters**

        * ``edge_id`` (str) - md5 hash edge id, if not provided, will be generated
          based on link attributes
        * ``label`` (str) - existing edge label
        * ``src_label`` (str) - existing edge source label
        * ``trgt_label`` (str) - existing edge target label
        * ``source`` (str) - existing edge source node id
        * ``target`` (str) - existing edge target node id
        * ``new_label`` (str) - new edge label
        * ``data`` (dict) - edge new data attributes
        * ``url`` (str) - edge new url attribute
        * ``style`` (str) - OS path to file or sting containing style to apply to edge
        * ``new_src_label`` (str) - new edge source label`
        * ``new_trgt_label`` (str) - new edge target label
        * ``src_label_style`` (str) - string with style to apply to source label
        * ``trgt_label_style`` (str) - strung with style to apply to target label

        Either of these must be provided to find link element to update:

        * ``edge_id`` MD5 hash or
        * ``label, source, target, src_label, trgt_label`` existing link attributes to calculate ``edge_id``

        ``edge_id`` calculated based on - ``label, source, target, src_label, trgt_label`` -
        attributes following this algorithm:

        1. Edge tuple produced: ``tuple(sorted([label, source, target, src_label, trgt_label]))``
        2. MD5 hash derived from tuple: ``hashlib.md5(",".join(edge_tup).encode()).hexdigest()``

        If no ``label, src_label, trgt_label`` provided, they substituted with empty values in
        assumption that values for existing link are empty as well.

        This method will replace existing or add new labels to the link.

        Existing data attribute will be amended with new values using dictionary
        like update method.

        New style will replace existing style.
        """
        data = data or {}
        link_data = {}
        # get new label
        new_label = new_label if new_label is not None else label
        new_src_label = new_src_label if new_src_label is not None else src_label
        new_trgt_label = new_trgt_label if new_trgt_label is not None else trgt_label
        # create edge id
        edge_tup = tuple(sorted([label, source, target, src_label, trgt_label]))
        new_edge_tup = tuple(
            sorted([new_label, source, target, new_src_label, new_trgt_label])
        )
        edge_id = (
            hashlib.md5(",".join(edge_tup).encode()).hexdigest()
            if not edge_id
            else edge_id
        )
        new_edge_id = (
            hashlib.md5(",".join(new_edge_tup).encode()).hexdigest()
            if not edge_id
            else edge_id
        )
        # update edge id
        if edge_id in self.edges_ids[self.current_diagram_id]:
            self.edges_ids[self.current_diagram_id].remove(edge_id)
            self.edges_ids[self.current_diagram_id].append(new_edge_id)
        else:
            log.warning(
                "update_link, link does not exist - source '{}', target '{}', label '{}'".format(
                    source, target, label
                )
            )
            return
        # find edge element
        edge = self.current_root.find("./object[@id='{}']".format(edge_id))
        # update labels and id
        edge.attrib.update({"id": new_edge_id, "label": new_label})
        if new_src_label:
            src_label_obj = self.current_root.find(
                "./mxCell[@id='{}']".format("{}-src".format(edge_id))
            )
            if not src_label_obj is None:
                src_label_obj.attrib.update(
                    {
                        "id": "{}-src".format(new_edge_id),
                        "value": new_src_label,
                        "parent": new_edge_id,
                    }
                )
                if src_label_style:
                    src_label_obj.attrib["style"] = src_label_style
            # create new src_label
            else:
                src_label_obj = ET.fromstring(
                    self.drawio_link_label_xml.format(
                        id="{}-src".format(new_edge_id),
                        label=new_src_label,
                        parent_id=new_edge_id,
                        style=src_label_style or self.default_link_label_style,
                        x="-0.5",
                        rel="1",
                    )
                )
                self.current_root.append(src_label_obj)
            kwargs["src_label"] = new_src_label
        if new_trgt_label:
            trgt_label_obj = self.current_root.find(
                "./mxCell[@id='{}']".format("{}-trgt".format(edge_id))
            )
            if not trgt_label_obj is None:
                trgt_label_obj.attrib.update(
                    {
                        "id": "{}-trgt".format(new_edge_id),
                        "value": new_trgt_label,
                        "parent": new_edge_id,
                    }
                )
                if trgt_label_style:
                    trgt_label_obj.attrib["style"] = trgt_label_style
            else:
                trgt_label_obj = ET.fromstring(
                    self.drawio_link_label_xml.format(
                        id="{}-trgt".format(new_edge_id),
                        label=new_trgt_label,
                        parent_id=new_edge_id,
                        style=trgt_label_style or self.default_link_label_style,
                        x="0.5",
                        rel="-1",
                    )
                )
                self.current_root.append(trgt_label_obj)
            kwargs["trgt_label"] = new_trgt_label
        # replace edge data and url
        link_data.update(data)
        link_data.update(kwargs)
        edge = self._add_data_or_url(edge, link_data, url)
        # update style
        mxCell_elem = edge.find("./mxCell")
        if os.path.isfile(style[:5000]):
            with open(style, "r") as style_file:
                mxCell_elem.attrib["style"] = style_file.read()
        elif style.strip():
            mxCell_elem.attrib["style"] = style

    def compare(
        self, data, diagram_name=None, missing_colour="#C0C0C0", new_colour="#00FF00"
    ):
        """
        Method to combine two graphs - existing and new - and produce resulting
        graph following these rules:

        * nodes and links present in new graph but not in existing graph considered
          as new and will be updated with ``new_colour`` style attribute by
          default highlighting them in green
        * nodes and links missing from new graph but present in existing graph considered
          as missing and will be updated with ``missing_colour`` style attribute
          by default highlighting them in gray
        * nodes and links present in both graphs will remain unchanged

        **Parameters**

        * ``data`` (dict) dictionary containing new graph data, dictionary format should be
          the same as for ``from_dict`` method.
        * ``missing_colour`` (str) colour to apply to missing elements
        * ``new_colour`` (str) colour to apply to new elements

        **Sample usage**::

            from N2G import drawio_diagram
            existing_graph = {
                "nodes": [
                    {"id": "node-1"},
                    {"id": "node-2"},
                    {"id": "node-3"}
            ],
                "links": [
                    {"source": "node-1", "target": "node-2", "label": "bla1"},
                    {"source": "node-2", "target": "node-3", "label": "bla2"},
                ]
            }
            new_graph = {
                "nodes": [
                    {"id": "node-99"},
                    {"id": "node-100", "style": "./Pics/router_1.txt", "width": 78, "height": 53},
                ],
                "links": [
                    {"source": "node-2", "target": "node-3", "label": "bla2"},
                    {"source": "node-99", "target": "node-3", "label": "bla99"},
                    {"source": "node-100", "target": "node-99", "label": "bla10099"},
                ]
            }
            drawing = drawio_diagram()
            drawing.from_dict(data=existing_graph)
            drawing.compare(new_graph)
            drawing.layout(algo="kk")
            drawing.dump_file(filename="compared_graph.drawio")

        """
        if diagram_name:
            self.go_to_diagram(diagram_name=diagram_name)
        else:
            self.go_to_diagram(diagram_index=0)
        if isinstance(data, dict):
            all_new_data_ids = set()
            new_elements = []
            # combine all edges under "links" key
            data.setdefault("links", [])
            data["links"] += data.pop("edges") if data.get("edges") else []
            # find new node elements and add them to graph
            for node in data.get("nodes", []):
                all_new_data_ids.add(node["id"])
                if not node["id"] in self.nodes_ids[self.current_diagram_id]:
                    new_elements.append(node["id"])
                    # add new node
                    self.add_node(**node)
            # find new link elements and add them to graph
            for link in data["links"]:
                link_tup = tuple(
                    sorted(
                        [
                            link.get("label", ""),
                            link["source"],
                            link["target"],
                            link.get("src_label", ""),
                            link.get("trgt_label", ""),
                        ]
                    )
                )
                id = hashlib.md5(",".join(link_tup).encode()).hexdigest()
                all_new_data_ids.add(id)
                if not id in self.edges_ids[self.current_diagram_id]:
                    new_elements.append(id)
                    # add new link
                    self.add_link(**link)
            # update new elements style attribute
            for id in new_elements:
                mxCell = self.current_root.find("./object[@id='{}']/mxCell".format(id))
                # convert style string to dictionary
                style_dict = {
                    i.split("=")[0]: i.split("=")[1]
                    for i in mxCell.attrib["style"].split(";")
                    if "=" in i
                }
                # update style colors
                if "fillColor" in style_dict:
                    style_dict["fillColor"] = new_colour
                else:
                    style_dict["strokeColor"] = new_colour
                style_dict["fontColor"] = new_colour
                # recreate style string
                mxCell.attrib["style"] = ";".join(
                    ["{}={}".format(k, v) for k, v in style_dict.items()]
                )
            # get all elements IDs set
            all_elements_ids = set(
                self.nodes_ids[self.current_diagram_id]
                + self.edges_ids[self.current_diagram_id]
            )
            # iterate over missing elements and update color for them
            for id in all_elements_ids.difference(all_new_data_ids):
                mxCell = self.current_root.find("./object[@id='{}']/mxCell".format(id))
                # convert style string to dictionary
                style_dict = {
                    i.split("=")[0]: i.split("=")[1]
                    for i in mxCell.attrib["style"].split(";")
                    if "=" in i
                }
                # update style colors
                if "fillColor" in style_dict:
                    style_dict["fillColor"] = missing_colour
                else:
                    style_dict["strokeColor"] = missing_colour
                style_dict["fontColor"] = missing_colour
                # recreate style string
                mxCell.attrib["style"] = ";".join(
                    ["{}={}".format(k, v) for k, v in style_dict.items()]
                )

    def delete_node(self, id=None, ids=None):
        """
        Method to delete node by its id. Bulk delete operation
        supported by providing list of node ids to delete.

        **Parameters**

        * ``id`` (str) id of single node to delete
        * ``ids`` (list) list of node ids to delete

        """
        ids = ids or []
        ids = ids + [id] if id else ids
        for node_id in ids:
            node = self.current_root.find("./object[@id='{}']".format(node_id))
            if not node is None:
                self.current_root.remove(node)
                self.nodes_ids[self.current_diagram_id].remove(node_id)
                # remove edges:
                # below xpath selects all parents - '..' - of children that
                # has source or target equal to node id
                for edge in self.current_root.iterfind(
                    ".//mxCell[@source='{}']/..".format(node_id)
                ):
                    self.edges_ids[self.current_diagram_id].remove(edge.get("id"))
                    self.current_root.remove(edge)
                for edge in self.current_root.iterfind(
                    ".//mxCell[@target='{}']/..".format(node_id)
                ):
                    self.edges_ids[self.current_diagram_id].remove(edge.get("id"))
                    self.current_root.remove(edge)

    def delete_link(self, id=None, ids=None, label="", source="", target="", **kwargs):
        """
        Method to delete link by its id. Bulk delete operation
        supported by providing list of link ids to delete.

        If link ``id`` or ``ids`` not provided, id calculated based on - ``label,
        source, target`` - attributes using this algorithm:

        1. Edge tuple produced: ``tuple(sorted([label, source, target]))``
        2. MD5 hash derived from tuple: ``hashlib.md5(",".join(edge_tup).encode()).hexdigest()``

        **Parameters**

        * ``id`` (str) id of single link to delete
        * ``ids`` (list) list of link ids to delete
        * ``label`` (str) link label to calculate id of single link to delete
        * ``source`` (str) link source to calculate id of single link to delete
        * ``target`` (str) link target to calculate id of single link to delete

        """
        ids = ids or []
        if not id and not ids:
            # create edge id
            src_label = kwargs.get("src_label", "")
            trgt_label = kwargs.get("trgt_label", "")
            edge_tup = tuple(sorted([source, target, label, src_label, trgt_label]))
            ids.append(hashlib.md5(",".join(edge_tup).encode()).hexdigest())
        else:
            ids = ids + [id] if id else ids
        for edge_id in ids:
            edge = self.current_root.find("./object[@id='{}']".format(edge_id))
            if not edge is None:
                self.current_root.remove(edge)
                self.edges_ids[self.current_diagram_id].remove(edge_id)



"""
V3D Diagrams Viewer
*******************

V3D Diagrams Viewer allows to start simple Flask WEB UI application
to visualize network data in 3D space using
`force-3d-graph <https://github.com/vasturiano/3d-force-graph>`_ library.

This viewer needs to have Flask installed::

    pip install flask

Flask installed as part of ``full`` extras as well.

To run using N2G CLI tool::

    N2G --v3d-viewer --diagram-file /Diagrams/sample_v3d_viewer_file.json

Access WEB UI application via URL ``http://127.0.0.1:9000`` using your browser.

By default Flask server starts and listens on all operating system interfaces, but
specific IP address and port number can be specified as required using ``--ip`` and
``--port`` N2G CLI arguments.

Where ``sample_v3d_viewer_file.json`` file content should contain JSON data
conforming to force-3d-graph input
`JSON syntax format <https://github.com/vasturiano/3d-force-graph#input-json-syntax>`_
for example::

    {
        "nodes": [
            {
                "id": "id1",
                "name": "name1",
                "val": 1
            },
            {
                "id": "id2",
                "name": "name2",
                "val": 10
            }
        ],
        "links": [
            {
                "source": "id1",
                "target": "id2"
            }
        ]
    }
"""
import os
import json
import logging

try:
    from flask import Flask, render_template, Markup

    HAS_FLASK = True
except ImportError:
    HAS_FLASK = False

log = logging.getLogger(__name__)

app_dir = os.path.abspath(os.path.dirname(__file__))

if HAS_FLASK:
    app = Flask(
        __name__, static_folder=app_dir, static_url_path="", template_folder=app_dir
    )
    # based on https://stackoverflow.com/a/19269087/12300761 answer
    app.jinja_env.filters["json"] = lambda v: Markup(json.dumps(v))

    @app.route("/")
    def home():
        return render_template("v3d_viewer.html", json_data=diagram_content)


def run_v3d_viewer(
    ip: str = "0.0.0.0",
    port: int = 9000,
    debug: bool = True,
    diagram_file: str = None,
    diagram_data: str = None,
    **kwargs,
) -> None:
    """
    Function to start Flask server to run V3D viewer application.

    :param ip: IP address to run server on
    :param port: port number to run server on
    :param debug: whether server should run in debug mode
    :param diagram_file: OS path to JSON file with diagram content
    :param diagram_data: JSON string with diagram content
    :param kwargs: any additional argument to pass to Flask ``app.run`` call
    """
    global diagram_content
    if diagram_file:
        log.debug(
            f"Starting server on 'http://{ip}:{port}', diagram file: '{diagram_file}'"
        )
        with open(diagram_file, mode="r", encoding="utf-8") as f:
            diagram_content = f.read()
    elif diagram_data:
        log.debug(f"Starting server on 'http://{ip}:{port}' using diagram data")
        diagram_content = diagram_data
    else:
        log.error("No diagram file or diagram data provided, stopping...")
        return
    app.run(host=ip, port=port, debug=debug, **kwargs)



from copy import deepcopy


def merge_dict(a: dict, b: dict, use_deepcopy: bool = False) -> dict:
    """
    Helper function to recursively merge two dictionaries. This function only
    merges lists and dictionaries, if "a" dictionary value is not a dictionary or
    a list, then this value left unchanged, even if same key exists in dictionary "b"


    :param a: first dictionary to merge
    :param b: second dictionary to merge
    :param use_deepcopy: if True merges deepcopies of a and b dictionaries
    """
    result = deepcopy(a) if use_deepcopy else a
    for bk, bv in b.items():
        av = result.get(bk)
        if isinstance(av, dict) and isinstance(bv, dict):
            result[bk] = merge_dict(av, bv)
        elif isinstance(av, list) and isinstance(bv, list):
            result[bk] = deepcopy(av) + deepcopy(bv) if use_deepcopy else av + bv
        elif isinstance(av, list) and isinstance(bv, dict):
            result[bk] = deepcopy(av) if use_deepcopy else av
            result[bk].append(deepcopy(bv) if use_deepcopy else bv)
        elif isinstance(av, dict) and isinstance(bv, list):
            result[bk] = deepcopy(bv) if use_deepcopy else bv
            result[bk].append(deepcopy(av) if use_deepcopy else av)
        elif bk not in result:
            result[bk] = deepcopy(bv) if use_deepcopy else bv
    return result



"""
JSON Data Plugin
****************

JSON data plugin loads structured data from JSON string and inputs it into
diagram class - if JSON string produces list, uses ``from_list`` method,
if JSON string produces dictionary uses ``from_dict`` method.

Sample Usage
------------

Code to demonstrate how to use ``json_data`` plugin::

    from N2G import v3d_diagramm
    from N2G import json_data

    sample_json_data = '''{
        "links": [{"data": {}, "label": "bla1", "source": "node-1", "src_label": "", "target": "node-2", "trgt_label": ""},
                {"data": {}, "label": "bla2", "source": "node-1", "src_label": "", "target": "node-3", "trgt_label": ""},
                {"data": {}, "label": "bla3", "source": "node-3", "src_label": "", "target": "node-5", "trgt_label": ""},
                {"data": {}, "label": "bla4", "source": "node-3", "src_label": "", "target": "node-4", "trgt_label": ""},
                {"data": {}, "label": "bla77", "source": "node-33", "src_label": "", "target": "node-44", "trgt_label": ""},
                {"data": {"cd": 123, "ef": 456}, "label": "bla6", "source": "node-6", "src_label": "", "target": "node-1", "trgt_label": ""}],
        "nodes": [{"color": "green", "data": {}, "id": "node-1", "label": "node-1", "nodeResolution": 16},
                {"color": "green", "data": {}, "id": "node-2", "label": "node-2", "nodeResolution": 8},
                {"color": "blue", "data": {"val": 4}, "id": "node-3", "label": "node-3", "nodeResolution": 8},
                {"color": "green", "data": {}, "id": "node-4", "label": "node-4", "nodeResolution": 8},
                {"color": "green", "data": {}, "id": "node-5", "label": "node-5", "nodeResolution": 8},
                {"color": "green", "data": {"a": "b", "c": "d"}, "id": "node-6", "label": "node-6", "nodeResolution": 8},
                {"color": "green", "data": {}, "id": "node-33", "label": "node-33", "nodeResolution": 8},
                {"color": "green", "data": {}, "id": "node-44", "label": "node-44", "nodeResolution": 8},
                {"color": "green", "data": {}, "id": "node-25", "label": "node-25", "nodeResolution": 8}]
    }'''

    v3d_drawing = create_v3d_diagram()
    json_data(v3d_drawing, sample_json_data)
    v3d_drawing.dump_file()

API Reference
-------------

.. autofunction:: N2G.plugins.data.json_data.json_data
"""
import logging
import json

# initiate logging
log = logging.getLogger(__name__)


def json_data(drawing, data):
    """
    Function to load graph data from JSON text.

    :param drawing: (obj) class object of one of N2G diagram plugins
    :param data: (str) JSON string to load

    If JSON string produces list, uses ``frm_list`` method, if dictionary uses ``from_dict`` method
    """
    loaded_data = json.loads(data)

    if isinstance(loaded_data, list):
        drawing.from_list(data=loaded_data)
    elif isinstance(loaded_data, dict):
        drawing.from_dict(data=loaded_data)
    else:
        raise TypeError(
            "N2G:json_data JSON data should load into list or dictionary, not '{}'".format(
                type(loaded_data)
            )
        )
